from typing import Callable
from unittest import mock

import datasets
import numpy as np
import pandas as pd
import pytest
import tensorflow as tf
import vaex

import dataquality
import dataquality as dq
from dataquality.exceptions import GalileoException, GalileoWarning
from dataquality.loggers.data_logger.base_data_logger import DataSet
from dataquality.loggers.data_logger.text_classification import (
    TextClassificationDataLogger,
)
from dataquality.schemas.split import Split


def test_duplicate_rows(set_test_config, cleanup_after_use) -> None:
    set_test_config(task_type="text_classification")
    text_inputs = [
        "what movies star bruce willis",
        "show me films with drew barrymore from the 1980s",
        "what movies starred both al pacino and robert deniro",
        "find me all of the movies that starred harold ramis and bill murray",
        "find me a movie with a quote about baseball in it",
    ]
    gold = ["A", "C", "B", "A", "C"]
    ids = [0, 1, 2, 3, 4]

    dq.log_data_samples(texts=text_inputs, labels=gold, split="validation", ids=ids)
    dq.log_data_samples(texts=text_inputs, labels=gold, split="training", ids=ids)

    with pytest.raises(GalileoException):
        dq.log_data_samples(texts=text_inputs, labels=gold, split="validation", ids=ids)

    dq.log_data_samples(texts=text_inputs, labels=gold, split="test", ids=ids)

    with pytest.raises(GalileoException):
        dq.log_data_samples(texts=text_inputs, labels=gold, split="training", ids=ids)


def test_duplicate_output_rows(set_test_config, cleanup_after_use) -> None:
    set_test_config(task_type="text_classification")
    text_inputs = [
        "what movies star bruce willis",
        "show me films with drew barrymore from the 1980s",
        "what movies starred both al pacino and robert deniro",
        "find me all of the movies that starred harold ramis and bill murray",
        "find me a movie with a quote about baseball in it",
    ]
    gold = ["A", "C", "B", "A", "C"]
    ids = list(range(5))

    dq.log_data_samples(texts=text_inputs, labels=gold, split="validation", ids=ids)
    dq.log_data_samples(texts=text_inputs, labels=gold, split="training", ids=ids)

    embs = np.random.rand(5, 100)
    logits = np.random.rand(5, 100)
    ids = list(range(5))
    dq.log_model_outputs(embs=embs, logits=logits, ids=ids, split="training", epoch=0)
    dq.log_model_outputs(embs=embs, logits=logits, ids=ids, split="training", epoch=0)

    with pytest.raises(GalileoException) as e:
        dq.get_data_logger().upload()

    assert str(e.value).startswith("It seems as though you do not have unique ids")


def test_log_data_sample(
    set_test_config: Callable, cleanup_after_use: Callable
) -> None:
    logger = TextClassificationDataLogger()
    with mock.patch("dataquality.core.log.get_data_logger") as mock_method:
        mock_method.return_value = logger
        dq.log_data_sample(text="sample 1", label="A", id=1, split="training")

        assert logger.texts == ["sample 1"]
        assert logger.labels == ["A"]
        assert logger.ids == [1]
        assert logger.split == Split.training


@pytest.mark.parametrize(
    "dataset",
    [
        pd.DataFrame(
            {
                "my_text": ["sample1", "sample2", "sample3"],
                "my_labels": ["A", "A", "B"],
                "my_id": [1, 2, 3],
            }
        ),
        vaex.from_dict(
            {
                "my_text": ["sample1", "sample2", "sample3"],
                "my_labels": ["A", "A", "B"],
                "my_id": [1, 2, 3],
            }
        ),
        [
            {"my_text": "sample1", "my_labels": "A", "my_id": 1},
            {"my_text": "sample2", "my_labels": "A", "my_id": 2},
            {"my_text": "sample3", "my_labels": "B", "my_id": 3},
        ],
        tf.data.Dataset.from_tensor_slices(
            {
                "my_text": ["sample1", "sample2", "sample3"],
                "my_labels": ["A", "A", "B"],
                "my_id": [1, 2, 3],
            }
        ),
        datasets.Dataset.from_dict(
            dict(
                my_text=["sample1", "sample2", "sample3"],
                my_labels=["A", "A", "B"],
                my_id=[1, 2, 3],
            )
        ),
    ],
)
def test_log_dataset(
    dataset: DataSet, set_test_config: Callable, cleanup_after_use: Callable
) -> None:
    logger = TextClassificationDataLogger()

    with mock.patch("dataquality.core.log.get_data_logger") as mock_method:
        mock_method.return_value = logger
        dq.log_dataset(
            dataset, text="my_text", label="my_labels", id="my_id", split="train"
        )

        assert logger.texts == ["sample1", "sample2", "sample3"]
        assert logger.labels == ["A", "A", "B"]
        assert logger.ids == [1, 2, 3]
        assert logger.split == Split.training


@pytest.mark.parametrize(
    "dataset",
    [
        [
            ("sample1", "A", "ID1"),
            ("sample2", "A", "ID2"),
            ("sample3", "B", "ID3"),
        ],
        tf.data.Dataset.from_tensor_slices(
            [
                ("sample1", "A", "ID1"),
                ("sample2", "A", "ID2"),
                ("sample3", "B", "ID3"),
            ]
        ),
    ],
)
def test_log_dataset_tuple(
    dataset: DataSet, set_test_config: Callable, cleanup_after_use: Callable
) -> None:
    logger = TextClassificationDataLogger()

    with mock.patch("dataquality.core.log.get_data_logger") as mock_method:
        mock_method.return_value = logger
        dq.log_dataset(dataset, text=0, label=1, id=2, split="train")

        assert logger.texts == ["sample1", "sample2", "sample3"]
        assert logger.labels == ["A", "A", "B"]
        assert logger.ids == ["ID1", "ID2", "ID3"]
        assert logger.split == Split.training


@pytest.mark.parametrize(
    "dataset",
    [
        [
            {"text": "sample1", "label": "A", "id": 1},
            {"text": "sample2", "label": "A", "id": 2},
            {"text": "sample3", "label": "B", "id": 3},
        ],
    ],
)
def test_log_dataset_default(
    dataset: DataSet, set_test_config: Callable, cleanup_after_use: Callable
) -> None:
    """Tests that the default keys work as expected when not passed in"""
    logger = TextClassificationDataLogger()
    with mock.patch("dataquality.core.log.get_data_logger") as mock_method:
        mock_method.return_value = logger
        dq.log_dataset(dataset, split="train")

        assert logger.texts == ["sample1", "sample2", "sample3"]
        assert logger.labels == ["A", "A", "B"]
        assert logger.ids == [1, 2, 3]
        assert logger.split == Split.training


@pytest.mark.parametrize(
    "dataset",
    [
        [
            {"text": "sample1", "label": "A", "id": 1},
            {"text": "sample2", "label": "A", "id": 2},
            {"text": "sample3", "label": "B", "id": 3},
        ],
    ],
)
def test_output_no_input_for_split(
    dataset: DataSet, set_test_config: Callable, cleanup_after_use: Callable
) -> None:
    """Validates that no error is thrown and a warning is printed on finish

    When you try to upload data for a split that has no input data
    """
    dq.log_dataset(dataset, split="train")
    dq.log_model_outputs(
        embs=np.random.rand(3, 50),
        logits=np.random.rand(3, 10),
        ids=[1, 2, 3],
        split="validation",
        epoch=0,
    )
    with pytest.warns(GalileoWarning, match=r"There was output data"):
        dq.get_data_logger().upload()


@pytest.mark.parametrize("set_labels_first", [True, False])
def test_logged_labels_dont_match_set_labels(
    set_labels_first: bool, set_test_config: Callable, cleanup_after_use: Callable
) -> None:
    """An error should be thrown when the set labels dont match the logged labels"""
    set_test_config(task_type="text_classification")
    labels = ["A", "B", "C"]
    # labels are the index, not the actual labels. No good
    dataset = [
        {"text": "sample1", "label": 1, "id": 1},
        {"text": "sample2", "label": 1, "id": 2},
        {"text": "sample3", "label": 2, "id": 3},
    ]
    if set_labels_first:
        dataquality.set_labels_for_run(labels)
        with pytest.raises(AssertionError) as e:
            dataquality.log_dataset(dataset, split="train")
    else:
        dataquality.log_dataset(dataset, split="train")
        dataquality.get_data_logger().logger_config.observed_num_labels = len(labels)
        dataquality.set_labels_for_run(labels)
        with pytest.raises(AssertionError) as e:
            dataquality.get_data_logger().validate_labels()
    assert str(e.value).startswith("Labels set to")


def test_log_int_labels(set_test_config: Callable, cleanup_after_use: Callable) -> None:
    """Tests that a user can log their labels as ints without issue"""
    set_test_config(task_type="text_classification")
    labels = [1, 2, 3]
    # labels are the index, not the actual labels. No good
    dataset = pd.DataFrame(
        [
            {"text": "sample1", "label": 1, "id": 1},
            {"text": "sample2", "label": 1, "id": 2},
            {"text": "sample3", "label": 2, "id": 3},
        ]
    )
    dataquality.log_dataset(dataset, split="train")
    dataquality.get_data_logger().logger_config.observed_num_labels = len(labels)

    dataquality.set_labels_for_run(labels)
    # Should not throw an error
    dataquality.get_data_logger().validate_labels()
    assert sorted(
        dataquality.get_data_logger().logger_config.observed_labels
    ) == sorted(["1", "2"])
    assert sorted(dataquality.get_data_logger().logger_config.labels) == sorted(
        ["1", "2", "3"]
    )



from enum import Enum
from typing import List, Union

from dataquality.exceptions import GalileoException


class Split(str, Enum):
    training = "training"
    train = "training"
    validation = "validation"
    test = "test"
    testing = "test"
    inference = "inference"

    @staticmethod
    def get_valid_attributes() -> List[str]:
        return list(map(lambda x: x.value, Split))


def conform_split(split: Union[str, Split]) -> Split:
    """Conforms split name to our naming conventions

    Raises GalileoException if split is invalid
    """
    if isinstance(split, Split):
        return split
    try:
        return Split[split]
    except KeyError:
        raise GalileoException(
            f"Split must be one of {Split.get_valid_attributes()} " f"but got {split}"
        )



import os
import threading
from typing import Dict, List

import h5py
import numpy as np
import vaex
from vaex.arrow.convert import arrow_string_array_from_buffers as convert_bytes
from vaex.dataframe import DataFrame

from dataquality.exceptions import GalileoException
from dataquality.loggers.base_logger import BaseLoggerAttributes
from dataquality.schemas.split import Split
from dataquality.utils import tqdm
from dataquality.utils.hdf5_store import HDF5_STORE, HDF5Store

lock = threading.Lock()


def _save_hdf5_file(location: str, file_name: str, data: Dict) -> None:
    """
    Helper function to save a dictionary as an hdf5 file
    """
    if not os.path.isdir(location):
        with lock:
            os.makedirs(location, exist_ok=True)
    file_path = f"{location}/{file_name}"
    with h5py.File(file_path, "w") as f:
        for col in data:
            group = f.create_group(f"/table/columns/{col}")
            col_data = np.array(data[col])
            if None in col_data:
                # h5py expects np.nan instead of None
                col_data = col_data.astype(np.float_)

            # String columns
            ctype = col_data.dtype
            if not np.issubdtype(ctype, np.number) and not np.issubdtype(
                ctype, np.bool_
            ):
                dtype = h5py.string_dtype()
                col_data = col_data.astype(dtype)
            else:
                dtype = col_data.dtype

            shape = col_data.shape
            group.create_dataset(
                "data", data=col_data, dtype=dtype, shape=shape, chunks=shape
            )


def _join_in_out_frames(in_df: DataFrame, out_df: DataFrame) -> DataFrame:
    """Helper function to join our input and output frames"""
    in_frame = in_df.copy()
    # There is an odd vaex bug where sometimes we lose the continuity of this dataframe
    # it's hard to reproduce, only shows up on linux, and hasn't been pinpointed yet
    # but materializing the join-key column fixes the issue
    # https://github.com/vaexio/vaex/issues/1972
    in_frame["id"] = in_frame["id"].values
    out_frame = out_df.copy()
    in_out = out_frame.join(
        in_frame, on="id", how="inner", lsuffix="_L", rsuffix="_R"
    ).copy()
    if len(in_out) != len(out_frame):
        num_missing = len(out_frame) - len(in_out)
        missing_ids = set(out_frame["id"].unique()) - set(in_out["id_L"].unique())
        split = out_frame["split"].unique()[0]
        raise GalileoException(
            "It seems there were logged outputs with no corresponding inputs logged "
            f"for split {split}. {num_missing} corresponding input IDs are missing:\n"
            f"{missing_ids}"
        )
    keep_cols = [c for c in in_out.get_column_names() if not c.endswith("_L")]
    in_out = in_out[keep_cols]
    for c in in_out.get_column_names():
        if c.endswith("_R"):
            in_out.rename(c, c.rstrip("_R"))
    return in_out


def validate_unique_ids(df: DataFrame, epoch_or_inf_name: str) -> None:
    """Helper function to validate the logged df has unique ids

    Fail gracefully otherwise
    """
    if df["id"].nunique() != len(df):
        epoch_or_inf_value, split = df[[epoch_or_inf_name, "split"]][0]
        dups = get_dup_ids(df)
        raise GalileoException(
            "It seems as though you do not have unique ids in this "
            f"split. Did you provide your own IDs? Or did you log for the same "
            f"split multiple times? Check where you are logging {split} data\n"
            f"split:{split}, {epoch_or_inf_name}: {epoch_or_inf_value}, "
            f"dup ids and counts: {dups}"
        )


def validate_ids_for_slice(df: DataFrame, col_name: str, value: str) -> None:
    """Helper function to validate a slice of the logged df has unique ids

    Slice data on the value of a vaex df column. This helper function only
    validates ids for slice equal to "epoch" or "inference_name".

    Fail gracefully on invalid ids or invalid slice
    """
    sliced_df = df[df[col_name] == value]
    if col_name == "inference_name":
        validate_unique_ids(sliced_df, "inference_name")
    elif col_name == "split":
        validate_unique_ids(sliced_df, "split")
    else:
        raise GalileoException(
            f"Invalid data slice. Column: {col_name}, Value: {value}"
        )


def validate_ids_for_df(df: DataFrame) -> None:
    """Helper function to validate the logged df has no duplicates

    For each split and inference name, validate that the sliced
    df has unique ids.

    Raises GalileoExcpetion on invalid df
    """
    for split in df["split"].unique():
        if split == Split.inference:
            inf_df = df[df["split"] == split]
            for inference_name in inf_df["inference_name"].unique():
                validate_ids_for_slice(inf_df, "inference_name", inference_name)
        else:
            validate_ids_for_slice(df, "split", split)


def get_dup_ids(df: DataFrame) -> List:
    """Gets the list of duplicate IDs in a dataframe, if any"""
    df_copy = df.copy()
    dup_df = df_copy.groupby(by="id", agg="count")
    return dup_df[dup_df["count"] > 1].to_records()


def _valid_prob_col(col: str) -> bool:
    return (
        col.endswith("id")
        or "gold" in col
        or "pred" in col
        or col.startswith("prob")
        or col.startswith("span")
        or col.startswith("data_")
        or col.startswith("galileo")
    )


def concat_hdf5_files(location: str, prob_only: bool) -> List[str]:
    """Concatenates all hdf5 in a directory using an HDF5 store

    Vaex stores a dataframe as an hdf5 file in a predictable format using groups

    Each column gets its own group, following "/table/columns/{col}/data

    We can exploit that by concatenating our datasets with that structure, so vaex
    can open the final file as a single dataframe

    :param location: The directory containing the files
    :param prob_only: If True, only the id, prob, and gold columns will be concatted
    """
    str_cols = []
    stores = {}
    files = os.listdir(location)
    df = vaex.open(f"{location}/{files[0]}")

    # Construct a store per column
    if prob_only:
        cols = [c for c in df.get_column_names() if _valid_prob_col(c)]
    else:
        cols = df.get_column_names()
    for col in cols:
        group = f"/table/columns/{col}/data"
        cval = df[col].to_numpy()
        if cval.ndim == 2:
            shape = cval[0].shape
        else:
            shape = ()
        dtype = df[col].dtype.numpy
        if not np.issubdtype(dtype, np.number) and not np.issubdtype(dtype, np.bool_):
            dtype = h5py.string_dtype(encoding="utf-8")
            str_cols.append(col)
        stores[col] = HDF5Store(f"{location}/{HDF5_STORE}", group, shape, dtype=dtype)

    for file in tqdm(files, leave=False, desc="Combining batches for upload"):
        fname = f"{location}/{file}"
        with h5py.File(fname, "r") as f:
            dset = f["table"]["columns"]
            keys = dset.keys()
            keys = [key for key in keys if key in cols]
            for key in keys:
                col_data = dset[key]
                # We have a string column, need to parse it
                if "indices" in col_data.keys():
                    assert key in str_cols, f"Unexpected string column ({key}) found"
                    indcs = col_data["indices"][:]
                    data = col_data["data"][:]
                    d = convert_bytes(data, indcs, None).to_numpy(zero_copy_only=False)
                else:
                    d = col_data["data"][:]
                if key in str_cols:
                    d = d.astype(h5py.string_dtype(encoding="utf-8"))
                stores[key].append(d)
        os.remove(fname)
    return str_cols


def drop_empty_columns(df: DataFrame) -> DataFrame:
    """Drops any columns that have no values"""
    cols = df.get_column_names()
    # Don't need to check the default columns, they've already been validated
    cols = [c for c in cols if c not in list(BaseLoggerAttributes)]
    col_counts = df.count(cols)
    empty_cols = [col for col, col_count in zip(cols, col_counts) if col_count == 0]
    return df.drop(*empty_cols) if empty_cols else df


def filter_df(df: DataFrame, col_name: str, value: str) -> DataFrame:
    """Filter vaex df on the value of a column

    Drop any columns for this df that are empty
    (e.g. metadata logged for a different split)
    """
    df_slice = df[df[col_name].str.equals(value)].copy()
    df_slice = drop_empty_columns(df_slice)
    # Remove the mask, work with only the filtered rows
    return df_slice.extract()


def rename_df(df: DataFrame, columns: Dict) -> DataFrame:
    """Renames a vaex df using a mapping"""
    df_copy = df.copy()
    for old, new in columns.items():
        df_copy.rename(old, new)
    return df_copy



import os
from time import sleep
from typing import Any, Dict, List, Optional, Tuple, Union

import requests
from pydantic.types import UUID4
from requests import Response

from dataquality.core._config import config, url_is_localhost
from dataquality.exceptions import GalileoException
from dataquality.schemas import RequestType, Route
from dataquality.schemas.dataframe import FileType
from dataquality.schemas.ner import TaggingSchema
from dataquality.schemas.split import conform_split
from dataquality.schemas.task_type import TaskType
from dataquality.utils.auth import headers


class ApiClient:
    def __check_login(self) -> None:
        if not config.token:
            raise GalileoException("You are not logged in. Call dataquality.login()")

    def _validate_response(self, res: Response) -> None:
        if not res.ok:
            msg = (
                "Something didn't go quite right. The api returned a non-ok status "
                f"code {res.status_code} with output: {res.text}"
            )
            raise GalileoException(msg)
        # If a run/split or a set of filters has no data, the API will return a 200
        # with a special header letting us know, so we don't try to load a file
        # with an unexpected format
        elif res.headers.get("Galileo-No-Data") == "true":
            msg = (
                "It seems there is no data for this request.\nEnsure you spelled "
                "everything correctly. Projects and runs are case sensitive."
            )
            raise GalileoException(msg)

    def _get_user_id(self) -> UUID4:
        self.__check_login()
        return self.get_current_user()["id"]

    def get_current_user(self) -> Dict:
        if not config.token:
            raise GalileoException(
                "Current user is not set! Please ensure GALILEO_USERNAME "
                "is set to the registered email"
            )

        return self.make_request(
            RequestType.GET, url=f"{config.api_url}/{Route.current_user}"
        )

    def valid_current_user(self) -> bool:
        try:
            self.get_current_user()
            return True
        except GalileoException:
            return False

    def make_request(
        self,
        request: RequestType,
        url: str,
        body: Dict = None,
        data: Dict = None,
        params: Dict = None,
        header: Dict = None,
    ) -> Any:
        """Makes an HTTP request.

        This is the center point of all functions and the main entry/exit for the
        dataquality client to interact with the server.
        """
        self.__check_login()
        header = header or headers(config.token)
        res = RequestType.get_method(request.value)(
            url, json=body, params=params, headers=header, data=data
        )
        self._validate_response(res)
        return res.json()

    def get_project(self, project_id: UUID4) -> Dict:
        return self.make_request(
            RequestType.GET,
            url=f"{config.api_url}/{Route.projects}/{project_id}",
        )

    def get_projects(self) -> List[Dict]:
        user_id = self._get_user_id()
        return self.make_request(
            RequestType.GET,
            url=f"{config.api_url}/{Route.users}/{user_id}/{Route.projects}",
        )

    def get_project_by_name(self, project_name: str) -> Dict:
        projs = self.make_request(
            RequestType.GET,
            url=f"{config.api_url}/{Route.projects}",
            params={"project_name": project_name},
        )
        return projs[0] if projs else {}

    def get_project_runs(self, project_id: UUID4) -> List[Dict]:
        """Gets all runs from a project by ID"""
        return self.make_request(
            RequestType.GET,
            url=f"{config.api_url}/{Route.projects}/{project_id}/{Route.runs}",
        )

    def get_project_runs_by_name(self, project_name: str) -> List[Dict]:
        """Gets all runs from a project by name"""
        proj = self.get_project_by_name(project_name)
        return self.make_request(
            RequestType.GET,
            url=f"{config.api_url}/{Route.projects}/{proj['id']}/{Route.runs}",
        )

    def get_project_run(self, project_id: UUID4, run_id: UUID4) -> Dict:
        """Gets a run in a project by ID"""
        return self.make_request(
            RequestType.GET,
            url=f"{config.api_url}/{Route.projects}/{project_id}/{Route.runs}/{run_id}",
        )

    def get_project_run_by_name(self, project_name: str, run_name: str) -> Dict:
        proj = self.get_project_by_name(project_name)
        if not proj:
            raise GalileoException(f"No project with name {project_name}")
        url = f"{config.api_url}/{Route.projects}/{proj['id']}/{Route.runs}"
        params = {"run_name": run_name}
        runs = self.make_request(RequestType.GET, url=url, params=params)
        return runs[0] if runs else {}

    def create_project(self, project_name: str, is_public: bool = True) -> Dict:
        """Creates a project given a name and returns the project information"""
        body = {"name": project_name, "is_public": is_public}
        return self.make_request(
            RequestType.POST, url=f"{config.api_url}/{Route.projects}", body=body
        )

    def create_run(self, project_name: str, run_name: str, task_type: TaskType) -> Dict:
        """Creates a run in a given project"""
        body = {"name": run_name, "task_type": task_type.value}
        proj = self.get_project_by_name(project_name)
        return self.make_request(
            RequestType.POST,
            url=f"{config.api_url}/{Route.projects}/{proj['id']}/{Route.runs}",
            body=body,
        )

    def reset_run(
        self, project_id: UUID4, run_id: UUID4, task_type: Optional[TaskType] = None
    ) -> None:
        """Resets a run by deleting the run with that name and creating a new one
        with the same name, getting a new UUID

        Called before any call to `dataquality.finish` if prior data was logged.
        see `dataquality.finish`
        """
        project_name = self.get_project(project_id)["name"]
        run = self.get_project_run(project_id, run_id)
        run_name = run["name"]
        task_type = task_type or TaskType.get_mapping(run["task_type"])

        # Delete the run
        self.delete_run(project_id, run_id)
        # Create a run with the same name
        new_run = self.create_run(project_name, run_name, task_type)
        # Update config
        config.current_run_id = new_run["id"]

    def delete_run(self, project_id: UUID4, run_id: UUID4) -> Dict:
        """Deletes a run

        This clears all metadata about the run, all object data, and the run itself
        """
        return self.make_request(
            RequestType.DELETE,
            url=f"{config.api_url}/{Route.projects}/{project_id}/{Route.runs}/{run_id}",
        )

    def delete_run_by_name(self, project_name: str, run_name: str) -> None:
        """Deletes a run via name

        This clears all metadata about the run, all object data, and the run itself
        """
        run = self.get_project_run_by_name(project_name, run_name)
        if not run:
            raise GalileoException(
                f"No project/run found with name {project_name}/{run_name}"
            )
        project_id = run["project_id"]
        run_id = run["id"]
        return self.make_request(
            RequestType.DELETE,
            url=f"{config.api_url}/{Route.projects}/{project_id}/{Route.runs}/{run_id}",
        )

    def delete_project(self, project_id: UUID4) -> Dict:
        """Deletes a project

        For each run in the project, this clears all metadata about the run,
        all object data, and the run itself
        """
        runs = self.get_project_runs(project_id)
        print("Deleting all runs within project.")
        for run in runs:
            print(f"Deleting run {run['name']}", end="... ")
            self.delete_run(project_id, run["id"])
            print("Done.")
        return self.make_request(
            RequestType.DELETE,
            url=f"{config.api_url}/{Route.projects}/{project_id}",
        )

    def delete_project_by_name(self, project_name: str) -> None:
        """Deletes a project by name

        For each run in the project, this clears all metadata about the run,
        all object data, and the run itself
        """
        project = self.get_project_by_name(project_name)
        if not project:
            raise GalileoException(f"No project found with name {project_name}")
        self.delete_project(project["id"])

    def _get_project_run_id(
        self, project_name: Optional[str] = None, run_name: Optional[str] = None
    ) -> Tuple[UUID4, UUID4]:
        """Helper function to get the project/run ids

        If project and run names are provided, the IDs will be fetched. Otherwise,
        This will return the currently initialized project/run IDs
        """
        if (project_name and not run_name) or (run_name and not project_name):
            raise GalileoException(
                "You must either provide both a project and run"
                "name or neither. If you provide neither, the "
                "currently initialized project/run will be used."
            )
        elif project_name and run_name:
            res = self.get_project_run_by_name(project_name, run_name)
            if not res:
                raise GalileoException(
                    f"No project/run found with name {project_name}/{run_name}"
                )
            project = res["project_id"]
            run = res["id"]
        else:
            project = config.current_project_id
            run = config.current_run_id
        return project, run

    def get_labels_for_run(
        self, project_name: str = None, run_name: str = None, task: str = None
    ) -> List[str]:
        """Gets the labels for a given run, else the currently initialized project/run

        If you do not provide a project and run name, the currently initialized
        project/run will be used. Otherwise you must provide both a project and run name
        If the run is a multi-label run, a task must be provided
        """
        if not task and config.task_type == TaskType.text_multi_label:
            raise GalileoException("For multi-label runs, a task name must be provided")

        project, run = self._get_project_run_id(
            project_name=project_name, run_name=run_name
        )

        url = f"{config.api_url}/{Route.content_path(project, run)}/{Route.labels}"
        params = {"task": task} if task else None
        res = self.make_request(RequestType.GET, url=url, params=params)
        return res["labels"]

    def get_tasks_for_run(
        self, project_name: str = None, run_name: str = None
    ) -> List[str]:
        """Gets the task names for a given multi-label run,

        If you do not provide a project and run name, the currently initialized
        project/run will be used. Otherwise you must provide both a project and run name

        This function is only valid for multi-label runs.
        """
        project, run = self._get_project_run_id(
            project_name=project_name, run_name=run_name
        )
        if self.get_task_type(project, run) != TaskType.text_multi_label:
            return []
        url = f"{config.api_url}/{Route.content_path(project, run)}/{Route.tasks}"
        res = self.make_request(RequestType.GET, url=url)
        return res["tasks"]

    def get_epochs_for_run(
        self, project_name: str, run_name: str, split: str
    ) -> List[int]:
        """Returns an ordered list of epochs for a run"""
        project, run = self._get_project_run_id(project_name, run_name)
        split = conform_split(split)
        url = (
            f"{config.api_url}/{Route.content_path(project, run, split)}/{Route.epochs}"
        )
        return self.make_request(RequestType.GET, url=url)

    def reprocess_run(
        self,
        project_name: str = None,
        run_name: str = None,
        labels: Union[List, List[List]] = None,
    ) -> Dict:
        """Reinitiate a project/run that has already been finished

        If a project and run name have been provided, that project/run will be
        reinitiated, otherwise we trigger the currently initialized project/run.

        This will clear out the current state in the server, and will recalculate
        * DEP score
        * UMAP Embeddings for visualization
        * Smart features

        :param project_name: If not set, will use the currently active project
        :param run_name: If not set, will use the currently active run
        :param labels: If set, will reprocess the run with these labels. If not set,
        labels will be used from the previously processed run. These must match the
        labels that were originally logged
        """
        project, run = self._get_project_run_id(project_name, run_name)
        project_name = project_name or self.get_project(project)["name"]
        run_name = run_name or self.get_project_run(project, run)["name"]

        # Multi-label has tasks and List[List] for labels
        if config.task_type == TaskType.text_multi_label:
            tasks = self.get_tasks_for_run(project_name, run_name)
            if not labels:
                labels = [
                    self.get_labels_for_run(project_name, run_name, t) for t in tasks
                ]
        else:
            tasks = []
            if not labels:
                try:
                    labels = self.get_labels_for_run(project_name, run_name)
                except GalileoException as e:
                    if "No data found" in str(e):
                        e = GalileoException(
                            f"It seems no data is available for run "
                            f"{project_name}/{run_name}"
                        )
                    raise e from None
                # There were no labels available for this run
                except KeyError:
                    raise GalileoException(
                        "It seems we cannot find the labels for this run. Please call "
                        "api_client.reprocess_run again, passing in your labels to the "
                        "'labels' keyword"
                    ) from None

        body = dict(
            project_id=str(project),
            run_id=str(run),
            labels=labels,
            tasks=tasks or None,
        )
        res = self.make_request(
            RequestType.POST, url=f"{config.api_url}/{Route.jobs}", body=body
        )
        print(
            f"Job {res['job_name']} successfully resubmitted. New results will be "
            f"available soon at {res['link']}"
        )
        return res

    def get_slice_by_name(self, project_name: str, slice_name: str) -> Dict:
        """Get a slice by name"""
        proj = self.get_project_by_name(project_name)
        url = f"{config.api_url}/{Route.content_path(proj['id'])}/{Route.slices}"
        params = {"slice_name": slice_name}
        slices = self.make_request(RequestType.GET, url=url, params=params)
        if not slices:
            raise GalileoException(
                f"No slice found for project {project_name} with name {slice_name}"
            )
        return slices[0]

    def get_metadata_columns(
        self, project_name: str, run_name: str, split: str
    ) -> Dict:
        """Lists the available metadata columns for a run/split

        Structure of data is:
        [{
            "name": str
            "is_categorical": bool
            "unique_values": Optional[List]
            "max": Optional[float]
            "min": Optional[float]
        },...]
        :param project_name:
        :param run_name:
        :param split:
        """
        project, run = self._get_project_run_id(project_name, run_name)
        split = conform_split(split)
        url = f"{config.api_url}/{Route.content_path(project, run, split)}/meta/columns"
        return self.make_request(RequestType.GET, url)

    def get_task_type(self, project_id: UUID4, run_id: UUID4) -> TaskType:
        return TaskType.get_mapping(
            self.get_project_run(project_id, run_id)["task_type"]
        )

    def export_run(
        self,
        project_name: str,
        run_name: str,
        split: str,
        file_name: str,
        inference_name: str = "",
        slice_name: Optional[str] = None,
        include_cols: Optional[List[str]] = None,
        col_mapping: Optional[Dict[str, str]] = None,
        hf_format: bool = False,
        tagging_schema: Optional[TaggingSchema] = None,
        filter_params: Dict = None,
    ) -> None:
        """Export a project/run to disk as a file

        :param project_name: The project name
        :param run_name: The run name
        :param split: The split to export on
        :param file_name: The file name. Must end in a supported FileType
        :param inference_name: Required if split is inference. The name of the inference
            split to get data for.
        :param slice_name: The optional slice name to export. If selected, this data
        from this slice will be exported only.
        :param include_cols: List of columns to include in the export. If not set,
        all columns will be exported.
        :param col_mapping: Dictionary of renamed column names for export.
        :param hf_format: (NER only)
            Whether to export the dataframe in a HuggingFace compatible format
        :param tagging_schema: (NER only)
            If hf_format is True, you must pass a tagging schema
        :param filter_params: Filters to apply to the dataframe before exporting. Only
        rows with matching filters will be included in the exported data. If a slice

        """
        project, run = self._get_project_run_id(project_name, run_name)
        ext = os.path.splitext(file_name)[-1].lstrip(".")

        assert ext in list(FileType), f"File must be one of {list(FileType)}"
        split = conform_split(split)
        body: Dict[str, Any] = dict(
            include_cols=include_cols,
            col_mapping=col_mapping,
            file_type=ext,
            hf_format=hf_format,
            tagging_schema=tagging_schema,
        )
        body["filter_params"] = {}
        if slice_name:
            slice_ = self.get_slice_by_name(project_name, slice_name)
            body["filter_params"].update(slice_["logic"])

        if filter_params:
            body["filter_params"].update(filter_params)

        if self.get_task_type(project, run) == TaskType.text_multi_label:
            body["task"] = self.get_tasks_for_run(project_name, run_name)[0]

        params = {"inference_name": inference_name}
        url = (
            f"{config.api_url}/{Route.content_path(project, run, split)}/{Route.export}"
        )
        self._export_dataframe_request(url, body, params, file_name)

    def get_project_run_name(
        self, project_id: Optional[UUID4] = None, run_id: Optional[UUID4] = None
    ) -> Tuple[str, str]:
        """Gets the project/run name given project/run IDs, or based on the config's

        Current project and run IDs
        """
        if (project_id and not run_id) or (run_id and not project_id):
            raise GalileoException(
                "You must either provide both the project and run IDs or neither "
                "(using the currently active project/run)"
            )
        if project_id and run_id:
            pid, rid = project_id, run_id
        elif config.current_project_id and config.current_run_id:
            pid, rid = config.current_project_id, config.current_run_id
        else:
            raise GalileoException(
                "You must either provide a project and run name or call "
                "dataquality.init() to initialize a run"
            )
        pname = self.get_project(pid)["name"]
        rname = self.get_project_run(pid, rid)["name"]
        return pname, rname

    def get_run_status(
        self, project_name: Optional[str] = None, run_name: Optional[str] = None
    ) -> Dict[str, Any]:
        if url_is_localhost(config.api_url):
            raise GalileoException(
                "You cannot check run status when running the server locally"
            )
        pid, rid = self._get_project_run_id(
            project_name=project_name, run_name=run_name
        )
        url = f"{config.api_url}/{Route.content_path(pid, rid)}/{Route.jobs}/status"
        statuses = self.make_request(RequestType.GET, url)["statuses"]
        status = sorted(statuses, key=lambda row: row["timestamp"], reverse=True)
        return status[0] if status else {}

    def wait_for_run(
        self, project_name: Optional[str] = None, run_name: Optional[str] = None
    ) -> None:
        print("Waiting for job...")
        while True:
            status = self.get_run_status(project_name=project_name, run_name=run_name)
            if status.get("status") == "finished":
                print(f"Done! Job finished with status {status.get('status')}")
                return
            elif status.get("status") == "errored":
                raise GalileoException(
                    f"It seems your run failed with status "
                    f"{status.get('status')}, error {status.get('message')}"
                )
            elif not status or status.get("status") == "started":
                sleep(2)
            else:
                raise GalileoException(
                    f"It seems there was an issue with your job. Received "
                    f"an unexpected status {status}"
                )

    def get_presigned_url(
        self, project_id: str, method: str, bucket_name: str, object_name: str
    ) -> str:
        response = self.make_request(
            request=RequestType.GET,
            url=f"{config.api_url}/{Route.presigned_url}",
            params={
                "api_url": config.api_url,
                "bucket_name": bucket_name,
                "object_name": object_name,
                "method": method.upper(),
                "project_id": project_id,
            },
        )
        return response["url"]

    def get_run_summary(
        self,
        project_name: str,
        run_name: str,
        split: str,
        task: str = None,
        inference_name: str = None,
        filter_params: Dict = None,
    ) -> Dict:
        """Gets overall run summary, or summary of a filtered subset.

        Use filter_params to apply arbitrary filters on the dataframe, based on the
        filter schema:
        https://api.dev.rungalileo.io/redoc#tag/insights
        """
        project, run = self._get_project_run_id(project_name, run_name)
        split = conform_split(split)
        path = Route.content_path(project, run, split)
        url = f"{config.api_url}/{path}/{Route.summary}"
        params = {}
        if inference_name:
            params["inference_name"] = inference_name
        body = {
            "hard_easy_threshold": True,
            "task": task,
            "filter_params": filter_params or {},
        }
        return self.make_request(RequestType.POST, url, body=body, params=params)

    def get_run_metrics(
        self,
        project_name: str,
        run_name: str,
        split: str,
        task: str = None,
        inference_name: str = None,
        category: str = "gold",
        filter_params: Dict = None,
    ) -> Dict[str, List]:
        project, run = self._get_project_run_id(project_name, run_name)
        split = conform_split(split)

        all_meta = self.get_metadata_columns(project_name, run_name, split)
        categorical_meta = [i["name"] for i in all_meta["meta"] if i["is_categorical"]]
        avl_cols = categorical_meta + ["gold", "pred"]
        if category not in avl_cols:
            raise GalileoException(
                f"Category must be one of {avl_cols} for this run but got {category}"
            )

        path = Route.content_path(project, run, split)
        url = f"{config.api_url}/{path}/{Route.groupby}"
        params = {"groupby_col": category}
        if inference_name:
            params["inference_name"] = inference_name
        body = {"task": task, "filter_params": filter_params or {}}
        return self.make_request(RequestType.POST, url, body=body, params=params)

    def get_column_distribution(
        self,
        project_name: str,
        run_name: str,
        split: str,
        task: str = None,
        inference_name: str = None,
        column: str = "data_error_potential",
        filter_params: Dict = None,
    ) -> Dict[str, List]:
        project, run = self._get_project_run_id(project_name, run_name)
        split = conform_split(split)

        all_meta = self.get_metadata_columns(project_name, run_name, split)
        continuous_meta = [i["name"] for i in all_meta["meta"] if i["is_continuous"]]
        avl_cols = continuous_meta + ["data_error_potential"]
        if column not in avl_cols:
            raise GalileoException(
                f"Column must be one of continuous columns {avl_cols} for this run "
                f"but got {column}"
            )

        path = Route.content_path(project, run, split)
        url = f"{config.api_url}/{path}/{Route.distribution}"
        params = {"col": column}
        if inference_name:
            params["inference_name"] = inference_name
        body = {"task": task, "filter_params": filter_params or {}}
        return self.make_request(RequestType.POST, url, body=body, params=params)

    def get_xray_cards(
        self, project_name: str, run_name: str, split: str, inference_name: str = None
    ) -> List[Dict[str, str]]:
        """Queries API for xray cards for a run/split"""
        project, run = self._get_project_run_id(project_name, run_name)
        path = Route.content_path(project, run, split)
        url = f"{config.api_url}/{path}/{Route.xray}"
        params = {"inference_name": inference_name} if inference_name else None
        return self.make_request(RequestType.GET, url, params=params)

    def get_edits(
        self, project_name: str, run_name: str, split: str, inference_name: str = None
    ) -> List:
        """Gets all edits for a run/split"""
        project, run = self._get_project_run_id(project_name, run_name)
        split = conform_split(split)

        url = (
            f"{config.api_url}/{Route.content_path(project, run, split)}/{Route.edits}"
        )
        params = {"inference_name": inference_name} if inference_name else None
        return self.make_request(RequestType.GET, url, params=params)

    def export_edits(
        self,
        project_name: str,
        run_name: str,
        split: str,
        file_name: str,
        inference_name: str = None,
        include_cols: Optional[List[str]] = None,
        col_mapping: Optional[Dict[str, str]] = None,
        hf_format: bool = False,
        tagging_schema: Optional[TaggingSchema] = None,
    ) -> None:
        """Export the edits of a project/run/split to disk as a file

        :param project_name: The project name
        :param run_name: The run name
        :param split: The split to export on
        :param file_name: The file name. Must end in a supported FileType
        :param inference_name: Required if split is inference. The name of the inference
            split to get data for.
        :param include_cols: List of columns to include in the export. If not set,
        all columns will be exported.
        :param col_mapping: Dictionary of renamed column names for export.
        :param hf_format: (NER only)
            Whether to export the dataframe in a HuggingFace compatible format
        :param tagging_schema: (NER only)
            If hf_format is True, you must pass a tagging schema
        :param filter_params: Filters to apply to the dataframe before exporting. Only
        rows with matching filters will be included in the exported data. If a slice
        """
        edits = self.get_edits(project_name, run_name, split, inference_name)

        ext = os.path.splitext(file_name)[-1].lstrip(".")
        assert ext in list(FileType), f"File must be one of {list(FileType)}"

        body: Dict[str, Any] = dict(
            include_cols=include_cols,
            col_mapping=col_mapping,
            file_type=ext,
            hf_format=hf_format,
            tagging_schema=tagging_schema,
            edit_ids=[edit["id"] for edit in edits],
        )
        url = f"{config.api_url}/{Route.export_edits}"
        params = {"inference_name": inference_name}
        self._export_dataframe_request(url, body, params, file_name)

    def _export_dataframe_request(
        self, url: str, body: Dict, params: Dict, file_name: str
    ) -> None:
        with requests.post(
            url, json=body, stream=True, headers=headers(config.token), params=params
        ) as r:
            self._validate_response(r)
            with open(file_name, "wb") as f:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)



import warnings
from collections import defaultdict
from enum import Enum, unique
from typing import Any, DefaultDict, Dict, List, Optional, Tuple, Union

import numpy as np
from numpy import ndarray

from dataquality.loggers.logger_config.text_ner import text_ner_logger_config
from dataquality.loggers.model_logger.base_model_logger import BaseGalileoModelLogger
from dataquality.schemas import __data_schema_version__
from dataquality.schemas.ner import NERErrorType, TaggingSchema
from dataquality.schemas.split import Split
from dataquality.utils.dq_logger import get_dq_logger


@unique
class GalileoModelLoggerAttributes(str, Enum):
    gold_emb = "gold_emb"
    gold_spans = "gold_spans"
    gold_dep = "gold_dep"
    embs = "embs"
    pred_emb = "pred_emb"
    pred_spans = "pred_spans"
    pred_dep = "pred_dep"
    probs = "probs"
    logits = "logits"
    ids = "ids"
    # mixin restriction on str (due to "str".split(...))
    split = "split"  # type: ignore
    epoch = "epoch"
    dep_scores = "dep_scores"
    log_helper_data = "log_helper_data"
    inference_name = "inference_name"

    @staticmethod
    def get_valid() -> List[str]:
        return list(map(lambda x: x.value, GalileoModelLoggerAttributes))


class TextNERModelLogger(BaseGalileoModelLogger):
    """
    Class for logging model output data of Text NER models to Galileo.

    * embs: List[np.ndarray]: Each np.ndarray represents all embeddings of a given
    sample. These embeddings are from the tokenized text, and will align with the tokens
    in the sample. If you have 12 samples in the dataset, with each sample of 20 tokens
    in length, and an embedding vector of size 768, len(embs) will be 12, and
    np.ndarray.shape is (20, 768).

    * logits: List[np.ndarray]: The NER prediction logits from the model
    for each token. These outputs are from the tokenized text, and will align with
    the tokens in the sample. If you have 12 samples in the dataset, with each sample
    of 20 tokens in length, and observed_num_labels as 40, len(probs) will be 12,
    and np.ndarray.shape is (20, 40).

    * probs: Probabilities: List[np.ndarray]: deprecated, use logits

    * ids: List[int]: These IDs must align with the input
    IDs for each sample input. This will be used to join them together for analysis
    by Galileo.

    * split: The model training/test/validation split for the samples being logged

    ex: (see the data input example in the DataLogger for NER
    `dataquality.get_data_logger().doc()`
    .. code-block:: python

        # Logged with `dataquality.log_model_outputs`
        logits =
            [np.array([model(the), model(president), model(is), model(joe),
            model(bi), model(##den), model(<pad>), model(<pad>), model(<pad>)]),
            np.array([model(joe), model(bi), model(##den), model(addressed),
            model(the), model(united), model(states), model(on), model(monday)])]

        embs =
            [np.array([emb(the), emb(president), emb(is), emb(joe),
            emb(bi), emb(##den), emb(<pad>), emb(<pad>), emb(<pad>)]),
            np.array([emb(joe), emb(bi), emb(##den), emb(addressed),
            emb(the), emb(united), emb(states), emb(on), emb(monday)])]

        epoch = 0
        ids = [0, 1]  # Must match the data input IDs
        split = "training"
        dataquality.log_model_outputs(
            embs=embs, logits=logits, ids=ids, split=split, epoch=epoch
        )
    """

    __logger_name__ = "text_ner"
    logger_config = text_ner_logger_config

    def __init__(
        self,
        embs: List[np.ndarray] = None,
        probs: List[np.ndarray] = None,
        logits: List[np.ndarray] = None,
        ids: Union[List, np.ndarray] = None,
        split: str = "",
        epoch: Optional[int] = None,
        inference_name: Optional[str] = None,
    ) -> None:
        super().__init__(
            embs=embs,
            probs=probs,
            logits=logits,
            ids=ids,
            split=split,
            epoch=epoch,
            inference_name=inference_name,
        )
        # Explicit cast to List from parent
        self.embs: List[np.ndarray] = list(self.embs)
        self.logits: List[np.ndarray] = list(self.logits)
        self.probs: List[np.ndarray] = list(self.probs)

        # Calculated internally
        self.gold_emb: List[List[np.ndarray]] = []
        self.gold_spans: List[List[Dict]] = []
        self.gold_dep: List[List[float]] = []

        self.pred_emb: List[List[np.ndarray]] = []
        self.pred_spans: List[List[Dict]] = []
        self.pred_dep: List[List[float]] = []

        # Used for helper data, does not get logged
        self.log_helper_data: Dict[str, Any] = {}

    @staticmethod
    def get_valid_attributes() -> List[str]:
        """
        Returns a list of valid attributes that GalileoModelConfig accepts
        :return: List[str]
        """
        return GalileoModelLoggerAttributes.get_valid()

    def validate(self) -> None:
        """
        Validates that the current config is correct.
        * embs, probs, and ids must exist and be the same length
        :return:
        """
        get_dq_logger().info(
            "Validating the output log.", split=self.split, epoch=self.epoch
        )
        if len(self.logits):
            self.probs = self.convert_logits_to_probs(self.logits).tolist()
        elif len(self.probs):
            warnings.warn("Usage of probs is deprecated, use logits instead")

        embs_len = len(self.embs)
        probs_len = len(self.probs)
        ids_len = len(self.ids)

        self.ids = self._convert_tensor_ndarray(self.ids)

        assert all([embs_len, probs_len, ids_len]), (
            f"All of emb, probs, and ids for your logger must be set, but "
            f"got emb:{bool(embs_len)}, probs:{bool(probs_len)}, ids:{bool(ids_len)}"
        )

        assert embs_len == probs_len == ids_len, (
            f"All of emb, probs, and ids for your logger must be the same "
            f"length, but got (emb, probs, ids) -> ({embs_len}, {probs_len}, {ids_len})"
        )

        # We need to average the embeddings for the tokens within a span
        # so each span has only 1 embedding vector
        logged_sample_ids = []
        for sample_id, sample_emb, sample_prob in zip(self.ids, self.embs, self.probs):
            # This will return True if there was a prediction or gold span
            sample_emb = self._convert_tensor_ndarray(sample_emb)
            if self._process_sample(sample_id, sample_emb, sample_prob):
                logged_sample_ids.append(sample_id)

        self.ids = logged_sample_ids
        assert self.ids, (
            "No samples in this batch had any gold or prediction spans. Logging will "
            "be skipped"
        )

    def _process_sample(
        self, sample_id: int, sample_emb: np.ndarray, sample_prob: np.ndarray
    ) -> bool:
        """Processes a sample. Returns whether or not the sample should be logged

        A sample should be logged only if there was at least 1 prediction span or 1
        gold span
        """
        get_dq_logger().info("Processing a sample.", split=self.split, epoch=self.epoch)
        # To extract metadata about the sample we are looking at
        sample_key = self.logger_config.get_sample_key(Split(self.split), sample_id)

        # Unpadded length of the sample. Used to extract true predicted spans
        # which are padded by the model
        sample_token_len = self.logger_config.sample_length[sample_key]
        # Get prediction spans
        sample_pred_spans = self._extract_pred_spans(sample_prob, sample_token_len)
        # Get gold (ground truth) spans
        gold_span_tup = self.logger_config.gold_spans.get(sample_key, [])
        sample_gold_spans: List[Dict] = [
            dict(start=start, end=end, label=label)
            for start, end, label in gold_span_tup
        ]
        # If there were no golds and no preds for a sample, don't log this sample
        if not sample_pred_spans and not sample_gold_spans:
            return False

        gold_dep, pred_dep = self._calculate_dep_scores(
            sample_prob, sample_gold_spans, sample_pred_spans, sample_token_len
        )
        gold_emb = self._extract_span_embeddings(sample_gold_spans, sample_emb)
        pred_emb = self._extract_span_embeddings(sample_pred_spans, sample_emb)

        self.pred_spans.append(sample_pred_spans)
        self.gold_spans.append(sample_gold_spans)
        self.pred_dep.append(pred_dep)
        self.gold_dep.append(gold_dep)
        self.gold_emb.append(gold_emb)
        self.pred_emb.append(pred_emb)
        return True

    def _extract_span_embeddings(
        self, spans: List[Dict], emb: np.ndarray
    ) -> List[np.ndarray]:
        """Get the embeddings for each span, on a per-sample basis

        We take the average of the token embeddings per span and use that as the span
        level embedding
        """
        get_dq_logger().info(
            "Extracting span embeddings.", split=self.split, epoch=self.epoch
        )
        embeddings = []
        for span in spans:
            start = span["start"]
            end = span["end"]
            span_embeddings = emb[start:end, :]
            avg_span_embedding = span_embeddings.mean(axis=0)
            embeddings.append(avg_span_embedding)
        return embeddings

    def _extract_pred_spans(self, pred_prob: np.ndarray, sample_len: int) -> List[Dict]:
        """
        Extract prediction labels from probabilities, and generate pred spans

        If the schema is non-BIO, we just first convert them into BIO then extract spans
        """
        # use length of the tokens stored to strip the pads
        # Drop the spans post first PAD
        get_dq_logger().info(
            "Extracting pred spans.", split=self.split, epoch=self.epoch
        )
        argmax_indices: List[int] = np.array(pred_prob).argmax(axis=1)
        pred_sequence: List[str] = [
            self.logger_config.labels[x] for x in argmax_indices
        ][0:sample_len]

        if self.logger_config.tagging_schema == TaggingSchema.BIO:
            pred_spans = self._extract_pred_spans_bio(pred_sequence)
        elif self.logger_config.tagging_schema == TaggingSchema.BILOU:
            pred_spans = self._extract_pred_spans_bilou(pred_sequence)
        else:  # BIOES
            pred_spans = self._extract_pred_spans_bioes(pred_sequence)
        return pred_spans

    def _extract_pred_spans_bio(self, pred_sequence: List[str]) -> List[Dict]:
        """
        Converts the prediction sequences into prediction span tokens

        Final format is of {'start': int, 'end': int, 'label': str}

        The function looks for a token with a B tag, and then continues forward
        in the sequence looking for sequential I tags with the same label.

        example:

        pred_sequence = [
            'I-PER', 'B-PER', 'B-LOC', 'B-PER', 'B-PER', 'B-PER', 'B-PER', 'B-PER',
            'B-PER', 'I-PER', 'B-PER', 'B-PER', 'B-PER', 'B-PER', 'B-PER', 'B-PER',
            'I-LOC', 'I-PER', 'I-LOC', 'I-PER', 'B-PER', 'B-PER', 'B-PER', 'B-PER',
            'B-ORG', 'O', 'B-PER', 'I-LOC', 'B-PER', 'I-ORG'
        ]
        self._extract_pred_spans_bio(pred_sequence)
        >> [
            {'start': 1, 'end': 2, 'label': 'PER'},
            {'start': 2, 'end': 3, 'label': 'LOC'},
            {'start': 3, 'end': 4, 'label': 'PER'},
            {'start': 4, 'end': 5, 'label': 'PER'},
            {'start': 5, 'end': 6, 'label': 'PER'},
            {'start': 6, 'end': 7, 'label': 'PER'},
            {'start': 7, 'end': 8, 'label': 'PER'},
            {'start': 8, 'end': 10, 'label': 'PER'},
            {'start': 10, 'end': 11, 'label': 'PER'},
        ]
        """
        pred_spans = []
        total_b_count = 0
        idx = 0
        # Use a while loop so we can skip rows already scanned in the inner loop
        while idx < len(pred_sequence):
            token = pred_sequence[idx]
            next_idx = idx + 1

            if not token.startswith("B"):
                idx += 1
                continue

            # We've hit a prediction. Continue until it's invalid
            # Invalid means we hit a new B or O tag, or the next I tag has a
            # different label
            total_b_count += 1
            token_val, token_label = self._split_token(token)

            for next_tok in pred_sequence[idx + 1 :]:
                # next_tok == "I" and the label matches the current B label
                if (
                    next_tok.startswith("I")
                    and self._split_token(next_tok)[1] == token_label
                ):
                    next_idx += 1
                else:
                    break

            pred_spans.append({"start": idx, "end": next_idx, "label": token_label})
            idx = next_idx

        assert total_b_count == len(pred_spans)
        return pred_spans

    def _extract_pred_spans_bilou(self, pred_sequence: List[str]) -> List[Dict]:
        """BILOU is a special case for BIO.

        The presense of I in a sequence does not mean a
        presence of a span until an L is successfully predicted.
        """
        pred_spans = []
        total_b_count = 0
        idx = 0
        found_end = False  # Tracks if there was an end tag predicted for BIOES

        # Use a while loop so we can skip rows already scanned in the inner loop
        while idx < len(pred_sequence):
            token = pred_sequence[idx]
            next_idx = idx + 1

            if self._is_single_token(token):
                total_b_count += 1
                # hit a single token prediction, update and continue
                token_val, token_label = self._split_token(token)
                pred_spans.append({"start": idx, "end": next_idx, "label": token_label})
                idx += 1
                continue

            if not self._is_before_token(token):
                idx += 1
                continue

            # We've hit a prediction. Continue until it's invalid
            # Invalid means we hit a new B or O tag, or the next I tag has a
            # different label
            token_val, token_label = self._split_token(token)

            for next_tok in pred_sequence[idx + 1 :]:
                if self._is_in_token(next_tok, token_label):
                    next_idx += 1
                # next_tok == "L" and the label matches the current B label
                elif self._is_end_token(next_tok, token_label):
                    next_idx += 1
                    found_end = True
                    total_b_count += 1
                    break
                else:
                    break
            if found_end:
                pred_spans.append({"start": idx, "end": next_idx, "label": token_label})
            idx = next_idx
            found_end = False

        assert total_b_count == len(pred_spans)
        return pred_spans

    def _extract_pred_spans_bioes(self, pred_sequence: List[str]) -> List[Dict]:
        """BIOES is a special case for BIO.

        The presense of I in a sequence does not mean a presence of a span until
        an E is successfully predicted.
        """
        pred_spans = []
        total_b_count = 0
        idx = 0
        found_end = False  # Tracks if there was an end tag predicted for BIOES

        # Use a while loop so we can skip rows already scanned in the inner loop
        while idx < len(pred_sequence):
            token = pred_sequence[idx]
            next_idx = idx + 1

            if self._is_single_token(token):
                total_b_count += 1
                # hit a single token prediction , update and continue
                token_val, token_label = self._split_token(token)
                pred_spans.append({"start": idx, "end": next_idx, "label": token_label})
                idx += 1
                continue

            if not self._is_before_token(token):
                idx += 1
                continue

            # We've hit a prediction. Continue until it's invalid
            # Invalid means we hit a new B or O tag, or the next I tag has a
            # different label
            token_val, token_label = self._split_token(token)

            for next_tok in pred_sequence[idx + 1 :]:
                # next_tok == "I" and the label matches the current B label
                if self._is_in_token(next_tok, token_label):
                    next_idx += 1
                # next_tok == "E" and the label matches the current B label
                elif self._is_end_token(next_tok, token_label):
                    next_idx += 1
                    found_end = True
                    total_b_count += 1
                    break
                else:
                    break
            if found_end:
                pred_spans.append({"start": idx, "end": next_idx, "label": token_label})
            idx = next_idx
            found_end = False

        assert total_b_count == len(pred_spans)
        return pred_spans

    def _is_single_token(self, tok: str) -> bool:
        return tok.startswith("U") or tok.startswith("S")

    def _is_before_token(self, tok: str) -> bool:
        return tok.startswith("B")

    def _is_in_token(self, tok: str, label: str) -> bool:
        """We are inside a token if the token is an I tag and the label matches"""
        # next_tok == "I" and the label matches the current B label
        return tok.startswith("I") and self._split_token(tok)[1] == label

    def _is_end_token(self, tok: str, label: str) -> bool:
        """We are at the end of a token if L/E tag and the label matches"""
        return (tok.startswith("E") or tok.startswith("L")) and self._split_token(tok)[
            1
        ] == label

    def _split_token(self, tok: str) -> Tuple[str, str]:
        """Split the token value and label

        A token starts with a tag and has a label sepearated by "-"
        ie B-DOG or E-SOME_VAL or I-my-label
        but we only want to split on the first "-" incase the label itself has a "-"
        """
        tok_tag, tok_label = tok.split("-", maxsplit=1)
        return tok_tag, tok_label

    def _construct_gold_sequence(
        self, len_sequence: int, gold_spans: List[Dict]
    ) -> List[str]:
        """
        Using gold spans and tagging schema, construct the underlying gold sequence
        e.g. gold_spans = [{start=5, end=8, label="nothing"}]
        gold_sequence = [O, O, O, O, O, B-nothing, I-nothing, I-nothing, O, O] for BIO
        gold_sequence = [O, O, O, O, O, B-nothing, I-nothing, L-nothing, O, O] for BILOU
        """
        get_dq_logger().info(
            "Constructing a gold sequence", split=self.split, epoch=self.epoch
        )
        gold_sequence = ["O"] * len_sequence
        for span in gold_spans:
            start = span["start"]
            end = span["end"]
            label = span["label"]
            if self.logger_config.tagging_schema == TaggingSchema.BIO:
                gold_sequence[start:end] = [f"I-{label}"] * (end - start)
                gold_sequence[start] = f"B-{label}"
            elif self.logger_config.tagging_schema == TaggingSchema.BILOU:
                if end - start == 1:
                    gold_sequence[start] = f"U-{label}"
                else:
                    gold_sequence[start:end] = [f"I-{label}"] * (end - start)
                    gold_sequence[start] = f"B-{label}"
                    gold_sequence[end - 1] = f"L-{label}"
            elif self.logger_config.tagging_schema == TaggingSchema.BIOES:
                if end - start == 1:
                    gold_sequence[start] = f"S-{label}"
                else:
                    gold_sequence[start:end] = [f"I-{label}"] * (end - start)
                    gold_sequence[start] = f"B-{label}"
                    gold_sequence[end - 1] = f"E-{label}"
        return gold_sequence

    def _calculate_dep_score_across_spans(
        self, spans: List[Dict], dep_scores: List[float]
    ) -> List[float]:
        """Computes dep score for all spans in a sample

        spans: All spans in a given sample
        dep_scores: DEP scores for every token in a sample, so len(dep_scores) is
            the number of tokens in a sentence
        """
        get_dq_logger().info(
            "Calculating DEP across spans.", split=self.split, epoch=self.epoch
        )
        dep_score_per_span = []
        for span in spans:
            start = span["start"]
            end = span["end"]
            dep_score_per_span.append(max(dep_scores[start:end]))
        assert len(dep_score_per_span) == len(spans)
        return dep_score_per_span

    def _calculate_dep_scores(
        self,
        pred_prob: np.ndarray,
        gold_spans: List[Dict],
        pred_spans: List[Dict],
        sample_token_len: int,
    ) -> Tuple[List[float], List[float]]:
        """Calculates dep scores for every span in a sample

        Compute DEP score for each token using gold and predicted sequences.
        Extract DEP score for each span using max of these token-level scores
        :param gold_spans: gold spans for a sample
        :param pred_spans: predicted spans for a sample
        :param pred_prob: seq_len x num_labels probability predictions for
            every token in a sample
        :param sample_token_len:  Unpadded length of the sample. Used to extract true
            predicted spans which are padded by the model
        :return: The DEP score per-token for both the gold spans and pred spans
        """
        get_dq_logger().info("Calculating DEP.", split=self.split, epoch=self.epoch)
        pred_prob = np.array(pred_prob)
        label2idx = {l: i for i, l in enumerate(self.logger_config.labels)}
        argmax_indices: List[int] = pred_prob.argmax(axis=1).tolist()
        pred_sequence: List[str] = [
            self.logger_config.labels[x] for x in argmax_indices
        ][0:sample_token_len]
        gold_sequence = self._construct_gold_sequence(len(pred_sequence), gold_spans)
        # Store dep scores for every token in the sample
        dep_scores_tokens = []
        for idx, token in enumerate(gold_sequence):
            g_label_idx = label2idx[token]  # index of ground truth
            token_prob_vector = pred_prob[idx]
            ordered_prob_vector = token_prob_vector.argsort()
            # We want the index of the highest probability that IS NOT the true label
            if ordered_prob_vector[-1] == g_label_idx:
                # Take the second highest probability because the highest was the label
                second_idx = ordered_prob_vector[-2]
            else:
                second_idx = ordered_prob_vector[-1]
            aum = token_prob_vector[g_label_idx] - token_prob_vector[second_idx]
            dep = (1 - aum) / 2  # normalize aum to dep
            assert 1.0 >= dep >= 0.0, f"DEP score is out of bounds with value {dep}"
            dep_scores_tokens.append(dep)
        assert sample_token_len == len(
            dep_scores_tokens
        ), "misalignment between total tokens and DEP scores"

        # Compute dep score of each span, which is effectively max dep score across
        # all tokens in a span
        gold_dep = self._calculate_dep_score_across_spans(gold_spans, dep_scores_tokens)
        pred_dep = self._calculate_dep_score_across_spans(pred_spans, dep_scores_tokens)
        return gold_dep, pred_dep

    def _get_data_dict(self) -> Dict[str, Any]:
        """Format row data for storage with vaex/hdf5

        In NER, rows are stored at the span level, not the sentence level, so we
        will have repeating "sentence_id" values, which is OK. We will also loop
        through the data twice, once for prediction span information
        (one of pred_span, pred_emb, pred_dep per span) and once for gold span
        information (one of gold_span, gold_emb, gold_dep per span)
        """
        data: defaultdict = defaultdict(list)

        get_dq_logger().info("Getting data dict.", split=self.split, epoch=self.epoch)
        # Loop through samples
        num_samples = len(self.ids)
        for idx in range(num_samples):
            sample_id = self.ids[idx]
            gold_spans = self.gold_spans[idx]
            gold_embs = self.gold_emb[idx]
            gold_deps = self.gold_dep[idx]
            pred_spans = self.pred_spans[idx]
            pred_embs = self.pred_emb[idx]
            pred_deps = self.pred_dep[idx]

            # We want to dedup gold and prediction spans, as many will match on
            # index. When the index matches, the embeddings and dep score will too,
            # so we only log the span once and flag it as both gold and pred
            pred_span_inds = [(i["start"], i["end"]) for i in pred_spans]
            pred_spans_check = set(pred_span_inds)
            # Loop through the gold spans
            for gold_span, gold_dep, gold_emb in zip(gold_spans, gold_deps, gold_embs):
                data = self._construct_gold_span_row(
                    data, sample_id, gold_span, gold_emb, gold_dep
                )

                span_ind = (gold_span["start"], gold_span["end"])
                if span_ind in pred_spans_check:
                    # Remove element from preds so it's not logged twice
                    ind = pred_span_inds.index(span_ind)
                    ps = pred_spans.pop(ind)
                    pred_embs.pop(ind)
                    pred_deps.pop(ind)
                    pred_span_inds.pop(ind)

                    data["is_pred"].append(True)
                    data["pred"].append(ps["label"])

                    # If indices match and tag doesn't, error_type is wrong_tag
                    error_type = (
                        NERErrorType.wrong_tag.value
                        if gold_span["label"] != ps["label"]
                        else NERErrorType.none.value
                    )
                    data["galileo_error_type"].append(error_type)

                else:
                    data["is_pred"].append(False)
                    data["pred"].append("")
                    error_type = self._get_span_error_type(
                        pred_span_inds, span_ind
                    ).value
                    data["galileo_error_type"].append(error_type)

            # Loop through the remaining pred spans
            for pred_span, pred_emb, pred_dep in zip(pred_spans, pred_embs, pred_deps):
                data = self._construct_pred_span_row(
                    data, sample_id, pred_span, pred_emb, pred_dep, gold_spans
                )
        return data

    def _construct_gold_span_row(
        self,
        data: DefaultDict,
        sample_id: int,
        gold_span: Dict,
        gold_emb: np.ndarray,
        gold_dep: float,
    ) -> DefaultDict:
        span_ind = (gold_span["start"], gold_span["end"])
        data = self._construct_span_row(
            data, sample_id, span_ind[0], span_ind[1], gold_dep, gold_emb
        )
        data["is_gold"].append(True)
        data["gold"].append(gold_span["label"])
        return data

    def _construct_pred_span_row(
        self,
        data: DefaultDict,
        sample_id: int,
        pred_span: Dict,
        pred_emb: np.ndarray,
        pred_dep: float,
        gold_spans: List[Dict],
    ) -> DefaultDict:
        start, end = pred_span["start"], pred_span["end"]
        data = self._construct_span_row(data, sample_id, start, end, pred_dep, pred_emb)
        data["is_gold"].append(False)
        data["is_pred"].append(True)
        data["pred"].append(pred_span["label"])
        data["gold"].append("")
        # Pred only spans are known as "ghost" spans (hallucinated) or no error
        err = (
            NERErrorType.ghost_span.value
            if self._is_ghost_span(pred_span, gold_spans)
            else NERErrorType.none.value
        )
        data["galileo_error_type"].append(err)
        return data

    def _is_ghost_span(self, pred_span: Dict, gold_spans: List[Dict]) -> bool:
        """Returns if the span is a ghost span

        A ghost span is a prediction span that has no overlap with any gold span.
        A ghost span is a pred_span where either:
        1. pred_end <= gold_start
        or
        2. pred_start >= gold_end

        For all gold spans
        """
        for gold_span in gold_spans:
            pred_start, pred_end = pred_span["start"], pred_span["end"]
            gold_start, gold_end = gold_span["start"], gold_span["end"]
            is_ghost = pred_start >= gold_end or pred_end <= gold_start
            if not is_ghost:  # If we ever hit not ghost, we can fail fast
                return False
        return True

    def _construct_span_row(
        self, d: DefaultDict, id: int, start: int, end: int, dep: float, emb: ndarray
    ) -> DefaultDict:
        d["sample_id"].append(id)
        d["epoch"].append(self.epoch)
        d["split"].append(Split(self.split).value)
        d["data_schema_version"].append(__data_schema_version__)
        d["span_start"].append(start)
        d["span_end"].append(end)
        d["data_error_potential"].append(dep)
        d["emb"].append(emb)
        return d

    def _get_span_error_type(
        self, pred_spans: List[Tuple[int, int]], gold_span: Tuple[int, int]
    ) -> NERErrorType:
        """Determines the proper span error

        When indices don't match, the error is either span_shift or missed_label
        * span_shift: overlapping span start/end indices
        * missed_label: no overlap between span start/end indices

        In this function, we only look at spans that don't have pred/gold alignment,
        so the error can only be span_shift or missed_label.
        """
        get_dq_logger().info(
            "Getting span error type.", split=self.split, epoch=self.epoch
        )
        gold_start, gold_end = gold_span
        # We start by assuming missed_label (because it's the worst case)
        # and update if we see overlap
        error_type = NERErrorType.missed_label
        for pred_span in pred_spans:
            pred_start, pred_end = pred_span
            if (
                gold_start <= pred_start <= gold_end
                or pred_start <= gold_start <= pred_end
            ):
                error_type = NERErrorType.span_shift
                break
        return error_type

    def __setattr__(self, key: Any, value: Any) -> None:
        if key not in self.get_valid_attributes():
            raise AttributeError(
                f"{key} is not a valid attribute of {self.__logger_name__} logger. "
                f"Only {self.get_valid_attributes()}"
            )
        super().__setattr__(key, value)

    def convert_logits_to_probs(
        self, sample_logits: Union[List, np.ndarray]
    ) -> np.ndarray:
        """Converts logits to probs via softmax per sample"""
        # axis ensures that in a matrix of probs with dims num_samples x num_classes
        # we take the softmax for each sample
        get_dq_logger().info(
            "Converting logits to probs.", split=self.split, epoch=self.epoch
        )
        token_probs = []
        for token_logits in sample_logits:
            token_probs.append(super().convert_logits_to_probs(token_logits))
        return np.array(token_probs, dtype=object)



import warnings
from typing import Any, Dict, Optional, Sequence, Union

import pytorch_lightning as pl
from pytorch_lightning.callbacks import Callback
from pytorch_lightning.trainer.supporters import CombinedDataset
from pytorch_lightning.utilities.types import STEP_OUTPUT
from torch.utils.data import Dataset
from torch.utils.data.dataloader import DataLoader

from dataquality import config
from dataquality.exceptions import GalileoException
from dataquality.loggers.data_logger import BaseGalileoDataLogger
from dataquality.loggers.model_logger import BaseGalileoModelLogger
from dataquality.schemas.split import Split
from dataquality.utils.thread_pool import ThreadPoolManager


class DataQualityCallback(Callback):
    """
    The PyTorch Lightning Callback for Galileo's dataquality module. This module
    handles the logging of input data and model loggers to Galileo. It makes the
    following assumptions:
    * Your model class has an attribute containing a valid GalileoModelLogger
    * You have a DataSet that extends PyTorch's DataSet and has an attribute containing
    a valid GalileoDataLogger
    """

    def __init__(self) -> None:
        self.checkpoint_data = {
            "epoch_start": False,
            "epoch": 0,
            "train_log": False,
            "val_log": False,
            "test_log": False,
        }

    def on_load_checkpoint(
        self,
        trainer: "pl.Trainer",
        pl_module: "pl.LightningModule",
        callback_state: Any,
    ) -> None:
        self.checkpoint_data = callback_state

    def on_save_checkpoint(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule", checkpoint: Any
    ) -> Dict[str, Any]:
        return self.checkpoint_data.copy()

    def _log_input_data(
        self,
        split: Split,
        dataloader: Optional[Union[DataLoader, Sequence[DataLoader]]],
    ) -> None:
        #
        # 🔭 Logging Inputs with Galileo!
        #
        if dataloader is None:
            warnings.warn(f"No {split} dataset available. Cannot log to Galileo")
            return
        loaders = dataloader if isinstance(dataloader, Sequence) else [dataloader]
        for loader in loaders:
            dataset = loader.dataset
            if isinstance(dataset, CombinedDataset):
                datasets = dataset.datasets
                if isinstance(datasets, Dataset):
                    dataset = datasets
                elif isinstance(datasets, Sequence):
                    warnings.warn(
                        "Galileo currently supports logging one dataset. "
                        "The first dataset in your CombinedDataset will be "
                        "logged"
                    )
                    dataset = datasets[0]
                else:
                    warnings.warn(
                        f"Your {split} dataset is of an unsupported type "
                        f"{type(datasets)}. Currently a single dataset is "
                        f"supported for logging."
                    )
                    return
            try:
                logger_attr = BaseGalileoDataLogger.get_data_logger_attr(dataset)
            except AttributeError:
                warnings.warn(
                    "No GalileoDataLogger found in your DataSet. Logging of input "
                    "data to Galileo will be skipped"
                )
                return

            data_logger: BaseGalileoDataLogger = getattr(dataset, logger_attr)
            data_logger.split = split
            try:
                data_logger.log()
            except GalileoException as e:
                warnings.warn(
                    f"Logging data inputs to Galileo could not be completed. See "
                    f"exception: {str(e)}"
                )
                return

    def _log_model_outputs(self, trainer: pl.Trainer, split: Split) -> None:
        try:
            logger_attr = BaseGalileoModelLogger.get_model_logger_attr(trainer.model)
        except AttributeError:
            warnings.warn(
                "No GalileoModelLogger found for this model, logging of model "
                "to Galileo will be skipped."
            )
            return

        model_logger: BaseGalileoModelLogger = getattr(trainer.model, logger_attr)
        model_logger.epoch = self.checkpoint_data["epoch"]
        model_logger.split = split.value

        try:
            model_logger.log()
        except GalileoException as e:
            warnings.warn(
                f"Logging model outputs to Galileo could not occur. "
                f"See exception: {str(e)}"
            )
            return

    def on_init_start(self, trainer: "pl.Trainer") -> None:
        self.checkpoint_data["epoch"] = 0
        assert (
            config.current_project_id and config.current_run_id
        ), "You must initialize dataquality before invoking a callback!"

    def on_train_start(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        if not self.checkpoint_data["train_log"]:
            self._log_input_data(Split.training, trainer.train_dataloader)
            self.checkpoint_data["train_log"] = True

    def on_test_start(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        if not self.checkpoint_data["test_log"]:
            self._log_input_data(Split.test, trainer.test_dataloaders)
            self.checkpoint_data["test_log"] = True

    def on_validation_start(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        if not self.checkpoint_data["val_log"]:
            self._log_input_data(Split.validation, trainer.val_dataloaders)
            self.checkpoint_data["val_log"] = True

    def on_predict_start(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        self._log_input_data(Split.inference, trainer.predict_dataloaders)

    def on_epoch_start(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        self.checkpoint_data["start_of_new_epoch"] = True

    def on_epoch_end(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule"
    ) -> None:
        # We need to implement epoch counting like this because there is a
        # bug in pytorch where at the end of each epoch The on_epoch_end
        # callback is called twice. This makes sure we only count it once
        # https://github.com/PyTorchLightning/pytorch-lightning/issues/5007
        if self.checkpoint_data["start_of_new_epoch"]:
            self.checkpoint_data["epoch"] += 1
        self.checkpoint_data["start_of_new_epoch"] = False

    def on_train_batch_end(
        self,
        trainer: "pl.Trainer",
        pl_module: "pl.LightningModule",
        outputs: STEP_OUTPUT,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        self._log_model_outputs(trainer, Split.training)

    def on_validation_batch_end(
        self,
        trainer: "pl.Trainer",
        pl_module: "pl.LightningModule",
        outputs: Optional[STEP_OUTPUT],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        self._log_model_outputs(trainer, Split.validation)

    def on_test_batch_end(
        self,
        trainer: "pl.Trainer",
        pl_module: "pl.LightningModule",
        outputs: Optional[STEP_OUTPUT],
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        self._log_model_outputs(trainer, Split.test)

    def on_predict_batch_end(
        self,
        trainer: "pl.Trainer",
        pl_module: "pl.LightningModule",
        outputs: Any,
        batch: Any,
        batch_idx: int,
        dataloader_idx: int,
    ) -> None:
        self._log_model_outputs(trainer, Split.inference)

    def teardown(
        self,
        trainer: "pl.Trainer",
        pl_module: "pl.LightningModule",
        stage: Optional[str] = None,
    ) -> None:
        # So we know everything has been fully written and finished
        ThreadPoolManager.wait_for_threads()



import hashlib
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
import spacy
import vaex
from spacy.language import Doc
from spacy.training import Example
from spacy.util import minibatch
from thinc.model import Model
from tqdm import tqdm
from vaex.dataframe import DataFrameLocal

import dataquality
from dataquality.integrations.spacy import log_input_examples, watch
from tests.conftest import SUBDIRS, TEST_PATH


def train_model(
    training_data: List[Tuple[str, Dict]],
    test_data: List[Tuple[str, Dict]],
    num_epochs: int = 5,
):
    nlp = spacy.blank("en")
    nlp.add_pipe("ner", last=True)
    minibatch_size = 3

    # Spacy pre-processing
    training_examples = []
    for text, annotations in training_data:
        doc = nlp.make_doc(text)
        training_examples.append(Example.from_dict(doc, annotations))

    optimizer = nlp.initialize(lambda: training_examples)

    # Galileo code
    watch(nlp)
    log_input_examples(training_examples, "training")
    log_input_examples(training_examples, "test")

    training_losses = []
    # with nlp.disable_pipes(*[pipe for pipe in nlp.pipe_names if pipe != "ner"]):
    for itn in range(num_epochs):
        print(f"Starting epoch {itn}")
        dataquality.set_epoch(itn)
        batches = minibatch(training_examples, minibatch_size)

        dataquality.set_split("training")
        for batch in tqdm(batches):
            training_loss = nlp.update(batch, drop=0.5, sgd=optimizer)
            training_losses.append(training_loss["ner"])
        print(training_loss["ner"])

        dataquality.set_split("test")
        nlp.evaluate(training_examples, batch_size=minibatch_size)

    # TODO: need to support the following line for inference
    # nlp('Thank you for your subscription renewal')

    # Evaluation Loop
    # TODO: Should change this with an actual evaluation loop

    logger = dataquality.get_data_logger(dataquality.config.task_type)
    logger.upload()

    return training_losses


def load_ner_data_from_local(
    split: str,
    epoch: int,
) -> (pd.DataFrame, DataFrameLocal, pd.DataFrame):
    """Loads post-logging locally created files.

    Returns: data, emb, and prob vaex dataframes
    """
    split_output_data = {}
    for subdir in SUBDIRS:
        file_path = (
            f"{TEST_PATH}/{split}/{epoch}/{subdir}/{subdir}."
            f"{'arrow' if subdir == 'data' else 'hdf5'}"
        )
        # Ensure files were cleaned up
        data = vaex.open(file_path)
        prob_cols = data.get_column_names(regex="prob*")
        for c in data.get_column_names():
            if c in prob_cols + ["emb"]:
                assert not np.isnan(data[c].values).any()
            else:
                vals = data[c].values
                assert all([i is not None and i != "nan" for i in vals])
        split_output_data[subdir] = data

    return (
        split_output_data["data"].to_pandas_df(),
        split_output_data["emb"],  # can't convert nested arrays to pandas df
        split_output_data["prob"].to_pandas_df(),
    )


def _calculate_emb_from_doc(doc: Doc) -> np.ndarray:
    hash_seed = int(hashlib.sha256(doc.text.encode("utf-8")).hexdigest(), 16) % 10**8
    np.random.seed(hash_seed)
    return np.random.rand(len(doc), 64)


class MockParserStepModel(Model):
    def __init__(self, docs: List[Doc]):
        self._func = self.mock_parser_step_model_forward

        tokvecs = []
        for doc in docs:
            tokvecs.extend(_calculate_emb_from_doc(doc))
        self.tokvecs = np.array(tokvecs)

    def mock_parser_step_model_forward(self, X, *args, **kwargs):
        """Returns logits that should construct an actual state"""


class MockTransitionBasedParserModel(Model):
    def __init__(self):
        self._func = self.mock_transition_based_parser_model_forward

    def mock_transition_based_parser_model_forward(self, X, *args, **kwargs):
        """Mocks TransitionBasedParser Model's forward func

        Returns a MockParserStepModel and a noop backprop_fn
        """
        # TODO: probably need to clean this up
        return MockParserStepModel(X), lambda d_output: 0.0
        # TODO: Does noop backprop_fn need to return some number?
        # I would guess it returns d_input



import warnings
from typing import Any, Union

import gorilla
from torch.nn import Module
from torch.utils.data import Dataset
from torch.utils.data.dataloader import DataLoader

from dataquality import config
from dataquality.exceptions import GalileoException
from dataquality.loggers.data_logger import BaseGalileoDataLogger
from dataquality.loggers.model_logger import BaseGalileoModelLogger
from dataquality.schemas.split import Split

_GORILLA_WATCH_SETTINGS = gorilla.Settings(allow_hit=True, store_hit=True)
_GORILLA_UNWATCH_SETTINGS = gorilla.Settings(allow_hit=True, store_hit=False)


def watch(model: Module) -> None:
    """
    Function to instantiate autologging to Galileo into vanilla Pytorch models. Because
    Pytorch doesn't support hooks or callbacks, we manually patch the apply function
    of the users Pytorch model, enabling us to force the logging to Galileo.

    NOTE: This will turn on autologging for ALL models of this class

    In order for this function to work, users must utilize the GalileoModelLogger in
    their Pytorch model class `__init__ or `forward` function, and implement a
    `forward` function

    :param model: The torch.nn.Module to watch
    :return: None
    """
    assert (
        config.current_project_id and config.current_run_id
    ), "You must initialize dataquality before invoking a callback!"
    if not isinstance(model, Module):
        raise GalileoException(
            f"Expected a pytorch model (torch.nn.Module). Received {str(type(model))}"
        )

    if not hasattr(model, "forward"):
        raise GalileoException(
            "Your model must implement a forward function in order "
            "to enable automated logging to Galileo!"
        )

    def patch_forward(cls: Module, *args: Any, **kwargs: Any) -> Any:
        """
        A patched forward function
        :param cls: The model class
        """
        # Run the forward for the model before logging
        orig = gorilla.get_original_attribute(cls, "forward")
        res = orig(*args, **kwargs)
        try:
            logger_attr = BaseGalileoModelLogger.get_model_logger_attr(cls)
        except AttributeError:  # User didn't specify a model logger
            warnings.warn(
                "Your model must utilize the GalileoModelLogger in order to enable "
                "automated logging to Galileo! Logging will be skipped."
            )
            return res
        model_logger: BaseGalileoModelLogger = getattr(cls, logger_attr)

        # Compare to None because 0 will be False
        if model_logger.epoch is None and model_logger.logger_config.cur_epoch is None:
            warnings.warn(
                "epoch must be set in your GalileoModelLogger for pytorch models to "
                "enable autologging to Galileo. If you are using Pytorch Lightning, "
                "consider using the DataQualityCallback in your trainer instead. "
                "You can also set the epoch using `dataquality.set_epoch`"
                "Logging will be skipped."
            )
            return res

        if not model_logger.split and not model_logger.logger_config.cur_split:
            if not model.training:
                warnings.warn(
                    "either split must be set in your GalileoModelLogger or your model "
                    "must be in 'training' mode (calling `model.train()`) for pytorch "
                    "models to enable autologging to Galileo. If you are using "
                    "Lightning consider using the DataQualityCallback in your trainer "
                    "You can also set the split using `dataquality.set_split`"
                    "instead."
                )
                return res
            else:
                warnings.warn(
                    "Model logger split was not set, but training mode is "
                    "set in your model. Using split=training for logging to "
                    "Galileo"
                )
                model_logger.split = "training"

        try:
            model_logger.log()
        except GalileoException as e:
            warnings.warn(
                f"Logging model outputs to Galileo could not be "
                f"completed. See exception: {str(e)}"
            )
        return res

    patch = gorilla.Patch(
        model.__class__, "forward", patch_forward, settings=_GORILLA_WATCH_SETTINGS
    )
    gorilla.apply(patch)


def unwatch(model: Module) -> None:
    """
    Unwatch a model that has previously been watched. This will turn off autologging
    for all models of this class

    :param model: The model to unwatch
    :return: None
    """
    if not hasattr(model, "forward"):
        return
    original = gorilla.get_original_attribute(model, "forward")
    gorilla.Patch(
        model.__class__, "forward", original, settings=_GORILLA_UNWATCH_SETTINGS
    )


def log_input_data(data: Union[DataLoader, Dataset], split: str) -> None:
    """
    Log input data to Galileo

    :param data: DataSet or DataLoader for training/validation/testing
    :param split: The data split. One of (training, validation, test, inference)
    :return: None
    """
    if split not in Split.get_valid_attributes():
        raise GalileoException(
            f"split must be one of {Split.get_valid_attributes()} but got {split}"
        )

    try:
        if isinstance(data, Dataset):
            logger_attr = BaseGalileoDataLogger.get_data_logger_attr(data)
            data_logger = getattr(data, logger_attr)
        elif isinstance(data, DataLoader):
            logger_attr = BaseGalileoDataLogger.get_data_logger_attr(data.dataset)
            data_logger = getattr(data.dataset, logger_attr)
        else:
            raise GalileoException(
                f"data must be one of (Dataset, DataLoader) but got {type(data)}"
            )
    except AttributeError:
        raise GalileoException(
            "Could not find a GalileoDataLogger as a part of your "
            "Dataset. You must include one to call this function."
        )
    data_logger.split = split
    data_logger.log()



from unittest import mock
from unittest.mock import MagicMock

import pytest

from dataquality import __version__
from dataquality.exceptions import GalileoException
from dataquality.utils import version
from tests.utils.mock_request import (
    mocked_healthcheck_request,
    mocked_healthcheck_request_new_api_version,
)


def test_version() -> None:
    assert __version__ is not None


def test_get_client_version() -> None:
    assert version._get_client_version() == __version__


@mock.patch("requests.get", side_effect=mocked_healthcheck_request)
def test_get_api_version(mock_get_api_version: MagicMock) -> None:
    assert version._get_api_version() == __version__


def test_parse_version_alpha() -> None:
    assert str(version._parse_version("v0.0.1-alpha")) == "0.0.1a0"


def test_parse_version_beta() -> None:
    assert str(version._parse_version("v0.0.2-beta")) == "0.0.2b0"


@mock.patch("requests.get", side_effect=mocked_healthcheck_request_new_api_version)
def test_version_check_fail(mock_get_healthcheck: MagicMock) -> None:
    with pytest.raises(GalileoException):
        version._version_check()


@mock.patch("requests.get", side_effect=mocked_healthcheck_request)
def test_version_check_pass(mock_get_healthcheck: MagicMock) -> None:
    version._version_check()
    assert True



import pytest

import dataquality


@pytest.fixture(autouse=True)
def reset_logger() -> None:
    # Only need to set one of data and models loggers since
    # they use the same logger config
    dataquality.get_data_logger().logger_config.reset()



import vaex

from dataquality.utils.vaex import drop_empty_columns


def test_drop_empty_cols() -> None:
    df1 = vaex.from_arrays(x=[1, 2, 3], y=[4, 5, 6], z=[7, 8, 9])
    df1["split"] = vaex.vconstant("training", length=len(df1))
    df2 = vaex.from_arrays(x=[1, 2, 3], y=[4, 5, 6], xz=[7, 8, 9])
    df2["split"] = vaex.vconstant("test", length=len(df2))

    in_frame = vaex.concat([df2, df1])
    split_df = in_frame[in_frame["split"] == "test"]
    assert split_df["z"].ismissing().to_numpy().all()
    split_df = drop_empty_columns(split_df)
    assert "z" not in split_df.get_column_names()



from itertools import chain
from typing import Callable, List
from unittest import mock

import datasets
import numpy as np
import pandas as pd
import pytest
import tensorflow as tf
import vaex

import dataquality
from dataquality.exceptions import GalileoException
from dataquality.loggers.data_logger.base_data_logger import DataSet
from dataquality.loggers.data_logger.text_ner import TextNERDataLogger
from dataquality.loggers.model_logger.text_ner import TextNERModelLogger
from dataquality.schemas.split import Split
from dataquality.schemas.task_type import TaskType
from dataquality.utils.thread_pool import ThreadPoolManager
from tests.conftest import TEST_PATH
from tests.utils.ner_constants import (
    GOLD_SPANS,
    LABELS,
    NER_INPUT_DATA,
    NER_INPUT_ITER,
    NER_INPUT_TUPLES,
    TEXT_INPUTS,
    TEXT_TOKENS,
)

model_logger = TextNERModelLogger()


def test_gold_span_extraction() -> None:
    logger = TextNERDataLogger()
    gold_spans = [
        {"start": 0, "end": 4, "label": "YEAR"},
        {"start": 17, "end": 29, "label": "ACTOR"},
    ]
    token_indices = [(0, 4), (5, 11), (12, 16), (17, 22), (17, 22), (23, 29), (23, 29)]
    new_spans = logger._extract_gold_spans(gold_spans, token_indices)
    good_new_spans = [
        {"start": 0, "end": 1, "label": "YEAR"},
        {"start": 3, "end": 7, "label": "ACTOR"},
    ]
    assert new_spans == good_new_spans


def test_construct_gold_sequence() -> None:
    len_sequence = 15
    case_1_seq_bio = [
        "O",
        "O",
        "B-PER",
        "I-PER",
        "B-ORG",
        "O",
        "B-MISC",
        "I-MISC",
        "I-MISC",
        "O",
        "B-PER",
        "B-ORG",
        "B-MISC",
        "O",
        "O",
    ]
    case_1_seq_bioes = [
        "O",
        "O",
        "B-PER",
        "E-PER",
        "S-ORG",
        "O",
        "B-MISC",
        "I-MISC",
        "E-MISC",
        "O",
        "S-PER",
        "S-ORG",
        "S-MISC",
        "O",
        "O",
    ]
    case_1_seq_bilou = [
        "O",
        "O",
        "B-PER",
        "L-PER",
        "U-ORG",
        "O",
        "B-MISC",
        "I-MISC",
        "L-MISC",
        "O",
        "U-PER",
        "U-ORG",
        "U-MISC",
        "O",
        "O",
    ]
    case_1_spans = [
        {"start": 2, "end": 4, "label": "PER"},
        {"start": 4, "end": 5, "label": "ORG"},
        {"start": 6, "end": 9, "label": "MISC"},
        {"start": 6, "end": 9, "label": "MISC"},
        {"start": 10, "end": 11, "label": "PER"},
        {"start": 11, "end": 12, "label": "ORG"},
        {"start": 12, "end": 13, "label": "MISC"},
    ]
    model_logger.logger_config.tagging_schema = "BIO"
    assert case_1_seq_bio == model_logger._construct_gold_sequence(
        len_sequence, case_1_spans
    )
    model_logger.logger_config.tagging_schema = "BIOES"
    assert case_1_seq_bioes == model_logger._construct_gold_sequence(
        len_sequence, case_1_spans
    )
    model_logger.logger_config.tagging_schema = "BILOU"
    assert case_1_seq_bilou == model_logger._construct_gold_sequence(
        len_sequence, case_1_spans
    )


def test_pred_span_extraction_bio() -> None:
    case_1_seq = ["O", "O", "B-PER", "I-ORG", "B-ORG", "O", "O"]
    case_1_spans = [
        {"start": 2, "end": 3, "label": "PER"},
        {"start": 4, "end": 5, "label": "ORG"},
    ]

    case_2_seq = ["O", "O", "B-PER", "I-PER", "I-ORG", "O", "O"]
    case_2_spans = [{"start": 2, "end": 4, "label": "PER"}]

    case_3_seq = ["O", "O", "B-PER", "I-ORG", "I-PER", "O", "O"]
    case_3_spans = [{"start": 2, "end": 3, "label": "PER"}]

    case_4_seq = ["O", "O", "B-PER", "B-ORG", "B-PER", "O", "B-PER"]
    case_4_spans = [
        {"start": 2, "end": 3, "label": "PER"},
        {"start": 3, "end": 4, "label": "ORG"},
        {"start": 4, "end": 5, "label": "PER"},
        {"start": 6, "end": 7, "label": "PER"},
    ]

    case_5_seq = ["O", "O", "B-PER", "I-PER", "I-PER", "O", "I-PER"]
    case_5_spans = [{"start": 2, "end": 5, "label": "PER"}]

    assert model_logger._extract_pred_spans_bio(case_1_seq) == case_1_spans
    assert model_logger._extract_pred_spans_bio(case_2_seq) == case_2_spans
    assert model_logger._extract_pred_spans_bio(case_3_seq) == case_3_spans
    assert model_logger._extract_pred_spans_bio(case_4_seq) == case_4_spans
    assert model_logger._extract_pred_spans_bio(case_5_seq) == case_5_spans


def test_pred_span_extraction_bioes() -> None:
    model_logger.logger_config.tagging_schema = "BIOES"
    case_1_seq = ["O", "O", "B-PER", "E-PER", "S-ORG", "O", "O"]
    case_1_spans = [
        {"start": 2, "end": 4, "label": "PER"},
        {"start": 4, "end": 5, "label": "ORG"},
    ]

    case_2_seq = ["O", "O", "B-PER", "I-PER", "I-ORG", "B-PER", "O"]
    case_2_spans = []

    case_3_seq = ["O", "O", "S-PER", "S-ORG", "B-PER", "O", "O"]
    case_3_spans = [
        {"start": 2, "end": 3, "label": "PER"},
        {"start": 3, "end": 4, "label": "ORG"},
    ]

    case_4_seq = ["O", "O", "B-PER", "B-ORG", "B-PER", "O", "B-PER"]
    case_4_spans = []

    case_5_seq = ["O", "O", "B-PER", "I-PER", "E-PER", "O", "I-PER"]
    case_5_spans = [{"start": 2, "end": 5, "label": "PER"}]

    assert model_logger._extract_pred_spans_bioes(case_1_seq) == case_1_spans
    assert model_logger._extract_pred_spans_bioes(case_2_seq) == case_2_spans
    assert model_logger._extract_pred_spans_bioes(case_3_seq) == case_3_spans
    assert model_logger._extract_pred_spans_bioes(case_4_seq) == case_4_spans
    assert model_logger._extract_pred_spans_bioes(case_5_seq) == case_5_spans


def test_pred_span_extraction_bilou() -> None:
    model_logger.logger_config.tagging_schema = "BILOU"
    case_1_seq = ["O", "O", "B-PER", "L-PER", "U-ORG", "O", "O"]
    case_1_spans = [
        {"start": 2, "end": 4, "label": "PER"},
        {"start": 4, "end": 5, "label": "ORG"},
    ]

    case_2_seq = ["O", "O", "B-PER", "I-PER", "I-ORG", "B-PER", "O"]
    case_2_spans = []

    case_3_seq = ["O", "O", "U-PER", "U-ORG", "B-PER", "O", "O"]
    case_3_spans = [
        {"start": 2, "end": 3, "label": "PER"},
        {"start": 3, "end": 4, "label": "ORG"},
    ]

    case_4_seq = ["O", "O", "B-PER", "B-ORG", "B-PER", "O", "B-PER"]
    case_4_spans = []

    case_5_seq = ["O", "O", "B-PER", "I-PER", "L-PER", "O", "I-PER"]
    case_5_spans = [{"start": 2, "end": 5, "label": "PER"}]

    assert model_logger._extract_pred_spans_bilou(case_1_seq) == case_1_spans
    assert model_logger._extract_pred_spans_bilou(case_2_seq) == case_2_spans
    assert model_logger._extract_pred_spans_bilou(case_3_seq) == case_3_spans
    assert model_logger._extract_pred_spans_bilou(case_4_seq) == case_4_spans
    assert model_logger._extract_pred_spans_bilou(case_5_seq) == case_5_spans


def test_calculate_dep_score_across_spans() -> None:
    dep_scores = [0.9, 0.2, 0.4, 0.3, 0.7, 0.9, 0.4, 0.3, 0.4, 0.3, 0.7, 0.9, 0.4, 0.3]
    spans = [
        {"start": 2, "end": 5, "label": "PER"},
        {"start": 6, "end": 10, "label": "MISC"},
        {"start": 12, "end": 13, "label": "ORG"},
    ]
    span_dep_score = [0.7, 0.4, 0.4]
    assert span_dep_score == model_logger._calculate_dep_score_across_spans(
        spans, dep_scores
    )


def test_calculate_dep_scores() -> None:
    model_logger.logger_config.tagging_schema = "BIO"
    pred_prob = np.array(
        [
            [0.9, 0.05, 0.05, 0, 0, 0, 0],
            [0.1, 0.7, 0.1, 0.1, 0, 0, 0],
            [0, 0, 0, 0.1, 0.9, 0, 0],
            [0.0, 0.4, 0.6, 0, 0, 0, 0],
            [0.2, 0.05, 0.05, 0, 0, 0.7, 0],
        ]
    )
    model_logger.logger_config.labels = [
        "B-PER",
        "I-PER",
        "B-ORG",
        "I-ORG",
        "O",
        "B-MISC",
        "I-MISC",
    ]
    sample_len = 5
    gold_spans = [
        {"start": 0, "end": 2, "label": "PER"},
        {"start": 4, "end": 5, "label": "MISC"},
    ]
    pred_spans = [
        {"start": 3, "end": 4, "label": "ORG"},
        {"start": 4, "end": 5, "label": "MISC"},
    ]
    gold_dep, pred_dep = model_logger._calculate_dep_scores(
        pred_prob, gold_spans, pred_spans, sample_len
    )
    assert gold_dep == [0.2, 0.25]
    assert pred_dep == [0.8, 0.25]


def test_ner_logging_bad_inputs(set_test_config: Callable, cleanup_after_use) -> None:
    set_test_config(task_type=TaskType.text_ner)
    dataquality.set_tagging_schema("BIO")

    labels = [
        "B-foo",
        "I-foo",
        "B-bar",
        "I-bar",
        "B-foo-bar",
        "I-foo-bar",
        "B-bar-foo",
        "I-bar-foo",
        "O",
    ]
    dataquality.set_labels_for_run(labels)

    text_inputs = [
        f"sample text for sentence {i}" for i in range(3)
    ]  # (tokens) sample te xt for sent en ce
    token_boundaries_all = [
        [(0, 6), (7, 11), (7, 11), (12, 15), (16, 24), (16, 24), (16, 24), (25, 26)]
        for _ in range(3)
    ]
    gold_spans = [
        [
            {"start": 7, "end": 11, "label": "foo"},
            {"start": 12, "end": 15, "label": "bar"},
            {"start": 16, "end": 27, "label": "bar-foo"},
        ],
        [{"start": 16, "end": 26, "label": "foo-bar"}],
        [],
    ]

    ids = [1, 2, 3]
    split = "training"

    # Handle spans that don't align with token boundaries
    with pytest.raises(AssertionError):
        dataquality.log_data_samples(
            texts=text_inputs,
            text_token_indices=token_boundaries_all,
            gold_spans=gold_spans,
            ids=ids,
            split=split,
        )

    # Handle labels that are in gold but missing in registered labels
    gold_spans = [
        [
            {"start": 7, "end": 11, "label": "foo"},
            {"start": 12, "end": 15, "label": "bar"},
            {"start": 16, "end": 26, "label": "bar-foo"},
        ],
        [{"start": 16, "end": 26, "label": "bad_label"}],
        [],
    ]
    # Handle spans that don't align with token boundaries
    with pytest.raises(AssertionError):
        dataquality.log_data_samples(
            texts=text_inputs,
            text_token_indices=token_boundaries_all,
            gold_spans=gold_spans,
            ids=ids,
            split=split,
        )


@pytest.mark.parametrize("as_tensor", [False, True])
def test_ner_logging(
    as_tensor: bool, cleanup_after_use: Callable, set_test_config: Callable
) -> None:
    """
    To validate:
    * dep scores are all 0 <= dep <= 1
    * assert correct start and end index for extracted spans
    * spans within gold cannot be nested
    * spans within pred cannot be nested
    * all spans should be either gold, pred, or both (never neither)
    * 3 rows
    """
    set_test_config(task_type=TaskType.text_ner)
    dataquality.set_tagging_schema("BIO")
    dataquality.set_split(Split.training)

    labels = [
        "B-foo",
        "I-foo",
        "B-bar",
        "I-bar",
        "B-foo-bar",
        "I-foo-bar",
        "B-bar-foo",
        "I-bar-foo",
        "O",
    ]
    dataquality.set_labels_for_run(labels)

    text_inputs = [
        f"sample text for sentence {i}" for i in range(3)
    ]  # (tokens) sample te xt for sent en ce
    token_boundaries_all = [
        [(0, 6), (7, 11), (7, 11), (12, 15), (16, 24), (16, 24), (16, 24), (25, 26)]
        for _ in range(3)
    ]
    gold_spans = [
        [
            {"start": 7, "end": 11, "label": "foo"},
            {"start": 12, "end": 15, "label": "bar"},
            {"start": 16, "end": 26, "label": "bar-foo"},
        ],
        [{"start": 16, "end": 26, "label": "foo-bar"}],
        [],
    ]

    ids = [0, 1, 2]
    split = "training"

    dataquality.log_data_samples(
        texts=text_inputs,
        text_token_indices=token_boundaries_all,
        gold_spans=gold_spans,
        ids=ids,
    )

    pred_prob = np.array(
        [
            [
                [0.0, 0.05, 0.05, 0, 0, 0, 0, 0, 0.9],
                [0.68, 0.1, 0.1, 0.1, 0, 0, 0.02, 0, 0],
                [0, 0.9, 0, 0.1, 0, 0, 0, 0, 0],
                [0.0, 0.4, 0.6, 0, 0, 0, 0, 0, 0],
                [0.0, 0, 0.05, 0, 0, 0, 0, 0, 0.95],
                [0.0, 0.05, 0.05, 0, 0, 0, 0, 0, 0.9],
                [0.0, 0.05, 0.05, 0, 0, 0, 0, 0, 0.9],
                [0.0, 0.05, 0.05, 0, 0, 0, 0.9, 0, 0],
            ],
            [
                [0.0, 0.05, 0.05, 0, 0, 0, 0, 0, 0.9],
                [0.0, 0.05, 0.05, 0, 0, 0, 0, 0, 0.9],
                [0.0, 0.05, 0.05, 0, 0, 0, 0, 0, 0.9],
                [0.0, 0.05, 0.05, 0, 0, 0, 0, 0, 0.9],
                [0.0, 0, 0.05, 0, 0.95, 0, 0, 0, 0.0],
                [0.0, 0.05, 0.05, 0, 0, 0.9, 0, 0, 0.0],
                [0.0, 0.05, 0.05, 0, 0, 0, 0, 0, 0.9],
                [0.0, 0.05, 0.05, 0, 0, 0, 0.9, 0, 0],
            ],
            [
                [0.0, 0.05, 0.05, 0, 0, 0, 0, 0, 0.9],
                [0.0, 0.05, 0.05, 0, 0, 0, 0.9, 0, 0],
                [0.2, 0.05, 0.05, 0, 0, 0, 0, 0.7, 0],
                [0.0, 0.05, 0.05, 0, 0, 0, 0, 0, 0.9],
                [0.0, 0.05, 0.05, 0.9, 0, 0, 0, 0, 0],
                [0.0, 0.05, 0.05, 0, 0, 0, 0, 0, 0.9],
                [0.0, 0.05, 0.05, 0, 0, 0, 0, 0, 0.9],
                [0.0, 0.05, 0.05, 0, 0, 0, 0, 0, 0.9],
            ],
        ]
    )

    dataquality.set_epoch(0)
    embs = np.random.rand(3, 8, 5)
    ids = [0, 1, 2]
    if as_tensor:
        embs = tf.convert_to_tensor(embs)
        pred_prob = tf.convert_to_tensor(pred_prob)
        ids = tf.convert_to_tensor(np.array(ids))
    dataquality.log_model_outputs(embs=embs, probs=pred_prob, ids=ids)

    ThreadPoolManager.wait_for_threads()
    c = dataquality.get_data_logger()
    c.validate_labels()
    c.upload()

    pred_spans = [
        [
            {"span_start": 1, "span_end": 3, "pred": "foo"},
            {"span_start": 3, "span_end": 4, "pred": "bar"},
            {"span_start": 7, "span_end": 8, "pred": "bar-foo"},
        ],
        [
            {"span_start": 4, "span_end": 6, "pred": "foo-bar"},
            {"span_start": 7, "span_end": 8, "pred": "bar-foo"},
        ],
        [{"span_start": 1, "span_end": 3, "pred": "bar-foo"}],
    ]

    gold_spans_correct = [
        [
            {"span_start": 1, "span_end": 3, "gold": "foo"},
            {"span_start": 3, "span_end": 4, "gold": "bar"},
            {"span_start": 4, "span_end": 8, "gold": "bar-foo"},
        ],
        [{"span_start": 4, "span_end": 8, "gold": "foo-bar"}],
        [],
    ]

    prob_path = f"{TEST_PATH}/{split}/0/prob/prob.hdf5"
    prob_df = vaex.open(prob_path)
    for i in prob_df.data_error_potential.to_numpy():
        assert 0 <= i <= 1

    for i in range(3):
        sample_pred_spans = pred_spans[i]
        pred_df = prob_df[prob_df[f"(is_pred) & (sample_id=={i})"]]
        df_pred_spans = pred_df[["span_start", "span_end", "pred"]].to_records()
        assert sample_pred_spans == df_pred_spans

        sample_gold_spans = gold_spans_correct[i]
        gold_df = prob_df[prob_df[f"(is_gold) & (sample_id=={i})"]]
        df_gold_spans = gold_df[["span_start", "span_end", "gold"]].to_records()
        assert sample_gold_spans == df_gold_spans

    assert len(prob_df[prob_df["(~is_gold) & (~is_pred)"]]) == 0

    emb_path = f"{TEST_PATH}/{split}/0/emb/emb.hdf5"
    emb_df = vaex.open(emb_path)

    assert len(emb_df) == len(prob_df)

    sample_path = f"{TEST_PATH}/{split}/0/data/data.arrow"
    sample_df = vaex.open(sample_path)

    assert len(sample_df) == 3

    # Test with logits
    c._cleanup()
    dataquality.set_labels_for_run(labels)
    dataquality.log_data_samples(
        texts=text_inputs,
        text_token_indices=token_boundaries_all,
        gold_spans=gold_spans,
        ids=ids,
        split=split,
    )
    dataquality.log_model_outputs(
        embs=embs,
        logits=pred_prob,
        ids=ids,
        split="training",
        epoch=0,
    )
    ThreadPoolManager.wait_for_threads()
    c = dataquality.get_data_logger()
    c.validate_labels()
    c.upload()
    prob_path = f"{TEST_PATH}/{split}/0/prob/prob.hdf5"
    prob_df = vaex.open(prob_path)
    for i in prob_df.data_error_potential.to_numpy():
        assert 0 <= i <= 1


def test_ghost_spans() -> None:
    gold_spans = [
        {"start": 7, "end": 11, "label": "foo"},  # span shift
        {"start": 12, "end": 15, "label": "bar"},  # wrong tag
        {"start": 16, "end": 26, "label": "bar-foo"},  # missed span
        {"start": 28, "end": 31, "label": "bar-foo"},  # span shift
    ]
    pred_spans = [
        {"start": 7, "end": 12, "label": "foo"},
        {"start": 12, "end": 15, "label": "bar-bar"},
        {"start": 27, "end": 30, "label": "foo"},
        {"start": 31, "end": 38, "label": "foo-foo"},  # ghost span
    ]
    results = [False, False, False, True]
    logger = TextNERModelLogger()
    for res, pred_span in zip(results, pred_spans):
        assert logger._is_ghost_span(pred_span, gold_spans) == res


def test_duplicate_rows(set_test_config, cleanup_after_use) -> None:
    set_test_config(task_type="text_ner")

    ids = list(range(5))

    dataquality.set_labels_for_run(LABELS)
    dataquality.set_tagging_schema("BIO")

    dataquality.log_data_samples(
        texts=TEXT_INPUTS,
        text_token_indices=TEXT_TOKENS,
        ids=ids,
        gold_spans=GOLD_SPANS,
        split="validation",
    )

    dataquality.log_data_samples(
        texts=TEXT_INPUTS,
        text_token_indices=TEXT_TOKENS,
        ids=ids,
        gold_spans=GOLD_SPANS,
        split="training",
    )

    with pytest.raises(GalileoException):
        dataquality.log_data_samples(
            texts=TEXT_INPUTS,
            text_token_indices=TEXT_TOKENS,
            ids=ids,
            gold_spans=GOLD_SPANS,
            split="validation",
        )

    dataquality.log_data_samples(
        texts=TEXT_INPUTS,
        text_token_indices=TEXT_TOKENS,
        ids=ids,
        gold_spans=GOLD_SPANS,
        split="test",
    )

    with pytest.raises(GalileoException):
        dataquality.log_data_samples(
            texts=TEXT_INPUTS,
            text_token_indices=TEXT_TOKENS,
            ids=ids,
            gold_spans=GOLD_SPANS,
            split="training",
        )


def test_duplicate_output_rows(set_test_config, cleanup_after_use) -> None:
    set_test_config(task_type="text_ner")

    ids = list(range(5))

    dataquality.set_labels_for_run(LABELS)
    dataquality.set_tagging_schema("BIO")

    dataquality.log_data_samples(
        texts=TEXT_INPUTS,
        text_token_indices=TEXT_TOKENS,
        ids=ids,
        gold_spans=GOLD_SPANS,
        split="validation",
    )

    dataquality.log_data_samples(
        texts=TEXT_INPUTS,
        text_token_indices=TEXT_TOKENS,
        ids=ids,
        gold_spans=GOLD_SPANS,
        split="training",
    )

    embs = [np.random.rand(119, 768) for _ in range(5)]
    logits = [np.random.rand(119, 28) for _ in range(5)]

    dataquality.log_model_outputs(
        embs=embs, logits=logits, ids=ids, split="training", epoch=0
    )
    dataquality.log_model_outputs(
        embs=embs, logits=logits, ids=ids, split="training", epoch=0
    )

    with pytest.raises(GalileoException) as e:
        dataquality.get_data_logger().upload()

    assert str(e.value).startswith("It seems as though you have duplicate spans")


def test_log_data_sample(
    set_test_config: Callable, cleanup_after_use: Callable
) -> None:
    set_test_config(task_type="text_ner")
    dataquality.set_labels_for_run(LABELS)
    dataquality.set_tagging_schema("BIO")
    logger = TextNERDataLogger()

    with mock.patch("dataquality.core.log.get_data_logger") as mock_method:
        mock_method.return_value = logger
        inp = {
            "text": TEXT_INPUTS[0],
            "text_token_indices": TEXT_TOKENS[0],
            "gold_spans": GOLD_SPANS[0],
            "id": 0,
            "split": "training",
        }
        dataquality.log_data_sample(**inp)

        assert logger.texts == [inp["text"]]
        assert logger.gold_spans == [inp["gold_spans"]]
        assert logger.ids == [0]
        assert logger.split == Split.training
        assert not hasattr(logger, "text_token_indices")  # Deleted after log
        flattened_tokens = list(chain(*inp["text_token_indices"]))
        assert logger.text_token_indices_flat == [flattened_tokens]


@pytest.mark.parametrize(
    "dataset",
    [
        pd.DataFrame(NER_INPUT_DATA),
        vaex.from_dict(NER_INPUT_DATA),
        NER_INPUT_ITER,
        datasets.Dataset.from_dict(NER_INPUT_DATA),
    ],
)
def test_log_dataset(
    dataset: DataSet, set_test_config: Callable, cleanup_after_use: Callable
) -> None:
    set_test_config(task_type="text_ner")
    dataquality.set_labels_for_run(LABELS)
    dataquality.set_tagging_schema("BIO")
    logger = TextNERDataLogger()

    with mock.patch("dataquality.core.log.get_data_logger") as mock_method:
        mock_method.return_value = logger
        dataquality.log_dataset(
            dataset,
            text="my_text",
            gold_spans="my_spans",
            id="my_id",
            text_token_indices="text_tokens",
            split="train",
        )

        assert logger.texts == TEXT_INPUTS
        assert logger.gold_spans == GOLD_SPANS
        # We delete it on successful log
        assert not hasattr(logger, "text_token_indices")
        flattened_input = [list(chain(*ind)) for ind in TEXT_TOKENS]
        assert logger.text_token_indices_flat == flattened_input
        assert logger.ids == list(range(5))
        assert logger.split == Split.training


@pytest.mark.parametrize(
    "dataset",
    [
        NER_INPUT_TUPLES,
    ],
)
def test_log_dataset_tuple(
    dataset: DataSet, set_test_config: Callable, cleanup_after_use: Callable
) -> None:
    set_test_config(task_type="text_ner")
    dataquality.set_labels_for_run(LABELS)
    dataquality.set_tagging_schema("BIO")
    logger = TextNERDataLogger()

    with mock.patch("dataquality.core.log.get_data_logger") as mock_method:
        mock_method.return_value = logger

        dataquality.log_dataset(
            dataset, text=0, gold_spans=1, id=2, text_token_indices=3, split="train"
        )

        assert logger.texts == TEXT_INPUTS
        assert logger.gold_spans == GOLD_SPANS
        # We delete it on successful log
        assert not hasattr(logger, "text_token_indices")
        flattened_input = [list(chain(*ind)) for ind in TEXT_TOKENS]
        assert logger.text_token_indices_flat == flattened_input
        assert logger.ids == list(range(5))
        assert logger.split == Split.training


@pytest.mark.parametrize(
    "labels, expected_err_message",
    [
        (LABELS, None),
        (
            [""],
            "No valid labels found among ['', 'NOT_']. A valid label is one that "
            "starts with either B-, I-, L-, E-, S-, or U- according to your "
            "particular tagging scheme",
        ),
        (
            ["B-"],
            "The class names following the tag cannot be empty. For example 'B-' is "
            "not allowed, but 'B-some_class' is.",
        ),
        (
            ["B_starting_word"],
            "No valid labels found among ['B_starting_word', 'NOT_B_starting_word']. "
            "A valid label is one that starts with either B-, I-, L-, E-, S-, or U- "
            "according to your particular tagging scheme",
        ),
    ],
)
def test_is_valid_span_label(
    set_test_config: Callable, labels: List[str], expected_err_message: str
):
    set_test_config(task_type="text_ner")

    dataquality.set_labels_for_run(labels)
    if expected_err_message:
        with pytest.raises(AssertionError) as e:
            TextNERDataLogger._clean_labels()
        assert e.value.args[0] == expected_err_message
    else:
        TextNERDataLogger._clean_labels()



from collections import defaultdict
from typing import DefaultDict, List, Optional, Set

import numpy as np
from pydantic import validator

from dataquality.loggers.logger_config.base_logger_config import BaseLoggerConfig


class TextMultiLabelLoggerConfig(BaseLoggerConfig):
    labels: Optional[List[List[str]]] = None
    observed_num_labels: Optional[List[int]] = None
    observed_labels: DefaultDict[int, Set] = defaultdict(set)
    tasks: Optional[List[str]] = None
    observed_num_tasks: int = 0
    binary: bool = False  # For binary multi label

    class Config:
        validate_assignment = True

    @validator("labels", always=True, pre=True)
    def clean_labels(cls, labels: List[List[str]]) -> List[List[str]]:
        cleaned_labels = []
        if isinstance(labels, np.ndarray):
            labels = labels.tolist()
        if labels is not None:
            assert isinstance(labels, List), "Labels must be a list of lists"
            for task_labels in labels:
                assert isinstance(task_labels, List), (
                    "Labels must be a list of lists. If you are running a binary "
                    "multi-label case, use `dq.set_tasks_for_run(labels, binary=True)"
                )
                if len(task_labels) == 1:
                    task_labels = [f"NOT_{task_labels[0]}", task_labels[0]]
                cleaned_labels.append([str(i) for i in task_labels])
        return cleaned_labels


text_multi_label_logger_config = TextMultiLabelLoggerConfig()



import pickle
from typing import Callable, Dict, List, Tuple
from unittest.mock import patch

import numpy as np
import pytest
import spacy
import vaex
from spacy.pipeline.ner import EntityRecognizer
from spacy.training import Example

import dataquality
from dataquality.exceptions import GalileoException
from dataquality.integrations.spacy import (
    GalileoEntityRecognizer,
    GalileoTransitionBasedParserModel,
    log_input_examples,
    unwatch,
    watch,
)
from dataquality.loggers.logger_config.text_ner import text_ner_logger_config
from dataquality.loggers.model_logger.text_ner import TextNERModelLogger
from dataquality.schemas.task_type import TaskType
from dataquality.utils.thread_pool import ThreadPoolManager
from tests.conftest import LOCATION
from tests.utils.spacy_integration import load_ner_data_from_local, train_model
from tests.utils.spacy_integration_constants import (
    LONG_SHORT_DATA,
    LONG_TRAIN_DATA,
    MISALIGNED_SPAN_DATA,
    NER_CLASS_LABELS,
    NER_TEST_DATA,
    NER_TRAINING_DATA,
    TestSpacyNerConstants,
)


def test_log_input_examples_without_watch(set_test_config, cleanup_after_use):
    text_ner_logger_config.reset()
    set_test_config(task_type=TaskType.text_ner)

    with pytest.raises(GalileoException) as e:
        log_input_examples(NER_TRAINING_DATA, split="training")
    assert (
        e.value.args[0]
        == "Galileo does not have any logged labels. Did you forget to call "
        "watch(nlp) before log_input_examples(...)?"
    )


def test_log_input_list_of_tuples(set_test_config, cleanup_after_use):
    text_ner_logger_config.reset()
    set_test_config(task_type=TaskType.text_ner)

    nlp = spacy.blank("en")
    nlp.add_pipe("ner")

    training_examples = []
    for text, annotations in NER_TRAINING_DATA:
        doc = nlp.make_doc(text)
        training_examples.append(Example.from_dict(doc, annotations))
    nlp.initialize(lambda: training_examples)

    watch(nlp)

    with pytest.raises(GalileoException) as e:
        log_input_examples(NER_TRAINING_DATA, "training")
    assert (
        e.value.args[0]
        == "Expected a <class 'spacy.training.example.Example'>. Received "
        "<class 'tuple'>"
    )


def test_log_input_examples(set_test_config, cleanup_after_use):
    text_ner_logger_config.reset()
    set_test_config(task_type=TaskType.text_ner)
    nlp = spacy.blank("en")
    nlp.add_pipe("ner")

    training_examples = []
    for text, annotations in NER_TRAINING_DATA:
        doc = nlp.make_doc(text)
        training_examples.append(Example.from_dict(doc, annotations))

    nlp.initialize(lambda: training_examples)

    watch(nlp)
    log_input_examples(training_examples, "training")

    # assert that we added ids to the examples for later joining with model outputs
    assert all(
        [
            examples.predicted.user_data["id"] == i
            for i, examples in enumerate(training_examples)
        ]
    )

    logged_data = vaex.open(f"{LOCATION}/input_data.arrow")

    assert logged_data["id"].tolist() == [0, 1, 2, 3, 4]
    assert logged_data["split"].tolist() == ["training"] * len(training_examples)
    assert all(
        [
            text == NER_TRAINING_DATA[i][0]
            for i, text in enumerate(logged_data["text"].tolist())
        ]
    )
    # Checks for logged gold spans matching converts from token to char idxs
    logged_token_indices = logged_data["text_token_indices"].tolist()
    for i, (split_plus_id, ents) in enumerate(
        dataquality.get_model_logger().logger_config.gold_spans.items()
    ):
        assert len(logged_token_indices[i]) % 2 == 0
        ents_as_char_idxs = []
        for ent in ents:
            start_char_idx = logged_token_indices[i][2 * ent[0]]
            end_char_idx = logged_token_indices[i][2 * ent[1] - 1]
            ents_as_char_idxs.append((start_char_idx, end_char_idx, ent[2]))

        assert NER_TRAINING_DATA[i][1]["entities"] == ents_as_char_idxs


def test_watch(set_test_config, cleanup_after_use):
    set_test_config(task_type=TaskType.text_ner)
    nlp = spacy.blank("en")
    nlp.add_pipe("ner")

    training_examples = []
    for text, annotations in NER_TRAINING_DATA:
        doc = nlp.make_doc(text)
        training_examples.append(Example.from_dict(doc, annotations))

    nlp.initialize(lambda: training_examples)
    watch(nlp)

    assert text_ner_logger_config.user_data["nlp"] == nlp
    assert dataquality.get_data_logger().logger_config.labels == NER_CLASS_LABELS
    assert dataquality.get_data_logger().logger_config.tagging_schema == "BILOU"

    assert isinstance(nlp.get_pipe("ner"), EntityRecognizer)
    assert isinstance(nlp.get_pipe("ner"), GalileoEntityRecognizer)


def test_unwatch(set_test_config):
    set_test_config(task_type=TaskType.text_ner)
    nlp = spacy.blank("en")
    original_ner = nlp.add_pipe("ner")

    training_examples = []
    for text, annotations in NER_TRAINING_DATA:
        doc = nlp.make_doc(text)
        training_examples.append(Example.from_dict(doc, annotations))

    nlp.initialize(lambda: training_examples)

    # This should be possible here
    pickle.dumps(nlp)

    for _ in range(3):  # This should be possible multiple times
        watch(nlp)
        unwatch(nlp)

    unwatched_ner = nlp.get_pipe("ner")
    assert isinstance(unwatched_ner, EntityRecognizer)
    assert not isinstance(unwatched_ner, GalileoEntityRecognizer)
    assert not isinstance(unwatched_ner.model, GalileoTransitionBasedParserModel)
    assert unwatched_ner.model == original_ner.model
    assert unwatched_ner.moves == original_ner.moves

    # Should be able to now save + load the pipeline component
    pickle.loads(pickle.dumps(nlp.get_pipe("ner")))
    # and the language
    pickle.loads(pickle.dumps(nlp))


def test_embeddings_get_updated(cleanup_after_use, set_test_config):
    """This test both checks our spacy wrapper end to end and that embs update.

    If embeddings stop updating that means the spacy architecture somehow changed
    and would make our user's embeddings seem meaningless
    """
    set_test_config(task_type=TaskType.text_ner)
    train_model(
        training_data=NER_TRAINING_DATA, test_data=NER_TRAINING_DATA, num_epochs=2
    )

    _, embs, _ = load_ner_data_from_local("training", epoch=1)
    embs = embs["emb"].to_numpy()

    dataquality.get_data_logger()._cleanup()

    train_model(
        training_data=NER_TRAINING_DATA, test_data=NER_TRAINING_DATA, num_epochs=1
    )

    _, embs_2, _ = load_ner_data_from_local("training", epoch=0)
    embs_2 = embs_2["emb"].to_numpy()

    assert embs.shape == embs_2.shape
    assert not np.allclose(embs, embs_2)


def test_spacy_ner(cleanup_after_use, set_test_config) -> None:
    """An end to end test of functionality"""
    spacy.util.fix_random_seed()
    set_test_config(task_type=TaskType.text_ner)
    num_epochs = 2
    training_losses = train_model(
        NER_TRAINING_DATA, NER_TEST_DATA, num_epochs=num_epochs
    )

    training_losses = np.array(training_losses).astype(np.float32)
    res = np.array(
        [25.50000334, 14.20009732, 41.1032235, 13.86421746], dtype=np.float32
    )
    assert np.allclose(training_losses, res, atol=1e-01)

    data, embs, probs = load_ner_data_from_local("training", epoch=num_epochs - 1)

    assert len(data) == 5
    assert all(data["id"] == range(len(data)))
    assert data.equals(
        TestSpacyNerConstants.gt_data
    ), f"Received the following data df {data}"

    assert embs["id"].tolist() == list(range(len(embs)))
    embs = embs["emb"].to_numpy().astype(np.float16)
    assert all([span_emb in TestSpacyNerConstants.gt_embs for span_emb in embs])

    # arrange the probs array to account for misordering of logged samples
    probs = (
        probs.sort_values(by=["sample_id", "span_start"]).drop("id", axis=1).round(4)
    )
    assert len(probs) == 8

    gt_probs = TestSpacyNerConstants.gt_probs
    for c in probs.columns:
        if np.issubdtype(probs[c].dtype, np.number):
            assert np.allclose(probs[c].values, gt_probs[c].values, atol=1e-3)
        else:
            assert (probs[c].values == gt_probs[c].values).all()


@pytest.mark.parametrize(
    "samples",
    [LONG_SHORT_DATA, LONG_TRAIN_DATA, NER_TRAINING_DATA],
)
def test_long_sample(
    samples: List[Tuple[str, Dict]],
    cleanup_after_use: Callable,
    set_test_config: Callable,
):
    """Tests logging a long sample during training"""
    TextNERModelLogger.logger_config.reset()
    set_test_config(task_type=TaskType.text_ner)

    nlp = spacy.blank("en")
    nlp.add_pipe("ner")
    all_examples = [
        Example.from_dict(nlp.make_doc(text), entities) for text, entities in samples
    ]
    optimizer = nlp.initialize(lambda: all_examples)

    old_log = TextNERModelLogger.log

    def new_log(*args, **kwargs):
        logger: TextNERModelLogger = args[0]
        assert len(logger.ids) == len(samples)
        assert len(logger.logits) == len(samples)
        assert len(logger.embs) == len(samples)

    TextNERModelLogger.log = new_log

    watch(nlp)
    log_input_examples(all_examples, split="training")

    dataquality.set_split("training")
    for epoch in range(2):
        dataquality.set_epoch(epoch)
        losses = {}
        nlp.update(all_examples, drop=0.5, sgd=optimizer, losses=losses)

    TextNERModelLogger.log = old_log
    ThreadPoolManager.wait_for_threads()
    del nlp


def test_inference_split_raises_warning(
    cleanup_after_use: Callable, set_test_config: Callable
) -> None:
    """Tests that inference mode raises a warning and continues without dq client"""
    TextNERModelLogger.logger_config.reset()
    set_test_config(task_type=TaskType.text_ner)

    nlp = spacy.blank("en")
    nlp.add_pipe("ner")
    all_examples = [
        Example.from_dict(nlp.make_doc(text), entities)
        for text, entities in NER_TRAINING_DATA
    ]
    nlp.initialize(lambda: all_examples)
    watch(nlp)
    dataquality.set_split(split="inference", inference_name="some_name")

    with patch(
        "dataquality.loggers.model_logger.text_ner.TextNERModelLogger"
    ) as mocked_model_logger_log:
        with pytest.warns(UserWarning) as record:
            nlp("some text here")
            assert len(record) == 1
            assert (
                record[0].message.args[0]
                == "Inference logging with Galileo coming soon. For now skipping "
                "logging"
            )
        assert not mocked_model_logger_log.called


def test_spacy_does_not_log_misaligned_entities(cleanup_after_use, set_test_config):
    TextNERModelLogger.logger_config.reset()
    set_test_config(task_type=TaskType.text_ner)

    nlp = spacy.blank("en")
    nlp.add_pipe("ner", last=True)

    def make_examples(data):
        examples = []
        for text, annotations in data:
            doc = nlp.make_doc(text)
            examples.append(Example.from_dict(doc, annotations))
        return examples

    nlp.initialize(lambda: make_examples(NER_TRAINING_DATA))

    training_examples = make_examples(MISALIGNED_SPAN_DATA)

    assert len(training_examples[0].reference.ents) == 0

    nlp.initialize(lambda: training_examples)

    # Galileo code
    watch(nlp)
    log_input_examples(training_examples, "training")

    logged_gold_spans = dataquality.get_data_logger().logger_config.gold_spans
    assert len(logged_gold_spans["training_0"]) == 0


@pytest.mark.parametrize(
    "training_data",
    [
        NER_TRAINING_DATA,
        [(text, {"entities": []}) for text, entities in NER_TRAINING_DATA],
        [
            (text, entities if i != 1 else {"entities": []})
            for i, (text, entities) in enumerate(NER_TRAINING_DATA)
        ],
    ],
)
def test_log_input_examples_have_no_gold_spans(
    set_test_config, cleanup_after_use, training_data
):
    TextNERModelLogger.logger_config.reset()
    set_test_config(task_type=TaskType.text_ner)
    nlp = spacy.blank("en")
    nlp.add_pipe("ner")

    def make_examples(data):
        examples = []
        for text, annotations in data:
            doc = nlp.make_doc(text)
            examples.append(Example.from_dict(doc, annotations))
        return examples

    nlp.initialize(lambda: make_examples(NER_TRAINING_DATA))

    training_examples = make_examples(training_data)

    watch(nlp)
    log_input_examples(training_examples, "training")

    samples_logged_gold_spans = dataquality.get_data_logger().logger_config.gold_spans

    for i, (_, logged_gold_spans) in enumerate(samples_logged_gold_spans.items()):
        original_spans = training_examples[i].reference.ents

        assert len(logged_gold_spans) == len(original_spans)
        for logged_gold_span, original_span in zip(logged_gold_spans, original_spans):
            assert logged_gold_span[0] == original_span.start
            assert logged_gold_span[1] == original_span.end
            assert logged_gold_span[2] == original_span.label_


def test_watch_nlp_with_no_gold_labels(set_test_config, cleanup_after_use):
    TextNERModelLogger.logger_config.reset()
    set_test_config(task_type=TaskType.text_ner)

    nlp = spacy.blank("en")
    nlp.add_pipe("ner")
    nlp.initialize()

    with pytest.raises(GalileoException) as e:
        watch(nlp)
    assert e.value.args[0] == (
        "Your nlp seems to not have been initialized with any "
        "ground truth spans. Galileo needs all labels to have "
        "been added to the model before calling "
        "watch(nlp). Please run `nlp.initialize(lambda: your_examples)` over a "
        "list of examples that include all of the gold spans you plan to make "
        "predictions over."
    )



from enum import Enum, unique
from typing import List


@unique
class TaskType(str, Enum):
    """Valid task types supported for logging by Galileo"""

    text_classification = "text_classification"
    text_multi_label = "text_multi_label"
    text_ner = "text_ner"

    @staticmethod
    def get_valid_tasks() -> List[str]:
        return list(map(lambda x: x.value, TaskType))

    @staticmethod
    def get_mapping(task_int: int) -> "TaskType":
        """Converts the servers task type enum to client names"""
        return {
            0: TaskType.text_classification,
            1: TaskType.text_multi_label,
            2: TaskType.text_ner,
        }[task_int]



import os
import shutil
from abc import abstractmethod
from enum import Enum, unique
from glob import glob
from typing import Any, List, Optional, Type, TypeVar, Union

import numpy as np

from dataquality.core._config import ConfigData, config
from dataquality.exceptions import GalileoException
from dataquality.loggers.logger_config.base_logger_config import (
    BaseLoggerConfig,
    base_logger_config,
)
from dataquality.schemas.split import Split, conform_split
from dataquality.schemas.task_type import TaskType
from dataquality.utils.dq_logger import upload_dq_log_file

try:
    from torch import Tensor

    TORCH_AVAILABLE = True
except ImportError:
    TORCH_AVAILABLE = False

try:
    import tensorflow as tf

    TF_AVAILABLE = True
except ImportError:
    TF_AVAILABLE = False

try:
    import datasets

    HF_AVAILABLE = True
except ImportError:
    HF_AVAILABLE = False


T = TypeVar("T", bound="BaseGalileoLogger")


@unique
class BaseLoggerAttributes(str, Enum):
    """A collection of all default attributes across all loggers"""

    texts = "texts"
    labels = "labels"
    ids = "ids"
    # mixin restriction on str (due to "str".split(...))
    split = "split"  # type: ignore
    meta = "meta"  # Metadata columns for logging
    prob = "prob"
    gold = "gold"
    embs = "embs"
    probs = "probs"
    logits = "logits"
    epoch = "epoch"
    data_error_potential = "data_error_potential"
    aum = "aum"
    text_tokenized = "text_tokenized"
    gold_spans = "gold_spans"
    pred_emb = "pred_emb"
    gold_emb = "gold_emb"
    pred_spans = "pred_spans"
    dep_scores = "dep_scores"
    text_token_indices = "text_token_indices"
    gold_dep = "gold_dep"
    pred_dep = "pred_dep"
    text_token_indices_flat = "text_token_indices_flat"
    log_helper_data = "log_helper_data"
    inference_name = "inference_name"

    @staticmethod
    def get_valid() -> List[str]:
        return list(map(lambda x: x.value, BaseLoggerAttributes))


class BaseGalileoLogger:
    """
    An abstract base class that all model logger and data loggers inherit from
    """

    __logger_name__ = ""
    LOG_FILE_DIR = f"{ConfigData.DEFAULT_GALILEO_CONFIG_DIR}/logs"
    logger_config: BaseLoggerConfig = base_logger_config

    def __init__(self) -> None:
        self.split: Optional[str] = None

    @staticmethod
    def get_valid_attributes() -> List[str]:
        return BaseLoggerAttributes.get_valid()

    @abstractmethod
    def validate(self) -> None:
        """Validates params passed in during logging. Implemented by child"""

    def set_split_epoch(self) -> None:
        if not self.split:
            if self.logger_config.cur_split:
                self.split = self.logger_config.cur_split
            else:
                raise GalileoException(
                    "You didn't log a split and did not set a split. Use "
                    "'dataquality.set_split' to set the split"
                )
        self.split = self.validate_split(self.split)
        # Set this config variable in validation, right before logging split data
        setattr(self.logger_config, f"{self.split}_logged", True)

    def is_valid(self) -> bool:
        try:
            self.validate()
        except AssertionError:
            return False
        return True

    @classmethod
    def non_inference_logged(cls) -> bool:
        """Return true if training, test, or validation data is logged

        If just inference data is logged then append data rather than overwriting.
        This flag is also used by the api to know which processing jobs to run.
        """
        return any(
            [
                cls.logger_config.training_logged,
                cls.logger_config.test_logged,
                cls.logger_config.validation_logged,
            ]
        )

    @abstractmethod
    def log(self) -> None:
        ...

    @staticmethod
    def _convert_tensor_ndarray(
        arr: Union[List, np.ndarray], attr: Optional[str] = None
    ) -> np.ndarray:
        """Handles numpy arrays and tensors conversions"""
        if TORCH_AVAILABLE:
            if isinstance(arr, Tensor):
                arr = arr.detach().cpu().numpy()
        if TF_AVAILABLE:
            if isinstance(arr, tf.Tensor):
                arr = arr.cpu().numpy()
        if isinstance(arr, np.ndarray):
            if attr == "Embedding":
                assert (
                    len(arr.shape) == 2
                ), f"{attr} tensor must be 2D shape, but got shape {arr.shape}"
            elif attr == "Prob":
                if config.task_type != TaskType.text_multi_label:
                    assert (
                        arr.ndim == 2
                    ), f"Probs/logits must have ndim=2, but got shape {arr.shape}"
                else:
                    # Because probs in multi label may not have a clear shape (each
                    # task may have a different number of probabilities
                    arr = np.array(arr, dtype=object)
                    assert arr.ndim > 1, (
                        f"Probs/logits must have at least 2 dimensions, "
                        f"they have {arr.ndim}"
                    )
            elif attr == "Labels" and config.task_type == TaskType.text_multi_label:
                arr = np.array(arr, dtype=object)
        return np.array(arr)

    @staticmethod
    def _convert_tensor_to_py(v: Any) -> Union[str, int]:
        """Converts pytorch and tensorflow tensors to strings or ints"""
        if isinstance(v, (int, str)):
            return v
        if TF_AVAILABLE:
            if isinstance(v, tf.Tensor):
                v = v.numpy()
                if isinstance(v, bytes):
                    v = v.decode("utf-8")
                else:
                    v = int(v)
        if TORCH_AVAILABLE:
            if isinstance(v, Tensor):
                v = int(v.numpy())  # Torch tensors cannot hold strings
        if not isinstance(v, (int, str)):
            raise GalileoException(
                f"Logged data should be of type int, string, pytorch tensor, "
                f"or tf tensor, but got {type(v)}"
            )
        return v

    @staticmethod
    def validate_task(task_type: Union[str, TaskType]) -> None:
        if task_type not in TaskType.get_valid_tasks():
            raise GalileoException(
                f"Task type {task_type} not valid. Choose one of "
                f"{TaskType.get_valid_tasks()}"
            )

    @classmethod
    def _cleanup(cls) -> None:
        """
        Cleans up the current run data locally
        """
        assert config.current_project_id
        assert config.current_run_id
        location = (
            f"{cls.LOG_FILE_DIR}/{config.current_project_id}"
            f"/{config.current_run_id}"
        )
        print("🧹 Cleaning up")
        for path in glob(f"{location}/*"):
            if os.path.isfile(path):
                os.remove(path)
            else:
                shutil.rmtree(path)

    @classmethod
    def upload(cls) -> None:
        ...

    @classmethod
    def get_all_subclasses(cls: Type[T]) -> List[Type[T]]:
        all_subclasses = []

        for subclass in cls.__subclasses__():
            all_subclasses.append(subclass)
            all_subclasses.extend(subclass.get_all_subclasses())

        return all_subclasses

    @classmethod
    def get_logger(cls: Type[T], task_type: TaskType) -> Type[T]:
        cls.validate_task(task_type)
        loggers = {i.__logger_name__: i for i in cls.get_all_subclasses()}
        return loggers[task_type]

    @classmethod
    def doc(cls) -> None:
        print(cls.__doc__)

    @classmethod
    def validate_split(cls, split: Union[str, Split]) -> str:
        return conform_split(split).value

    @classmethod
    def check_for_logging_failures(cls) -> None:
        """When a threaded logging call fails, it sets the logger_config.exception

        If that field is set, raise an exception here and stop the main process
        """
        # If a currently active thread crashed, check and raise a top level exception
        if cls.logger_config.exception:
            upload_dq_log_file()
            raise GalileoException(cls.logger_config.exception)

    @classmethod
    def is_hf_dataset(cls, df: Any) -> bool:
        if HF_AVAILABLE:
            return isinstance(df, datasets.Dataset)
        return False
